# -*- coding: utf-8 -*-
"""
Description
------------

Functions to deal with numpy arrays:

- :py:func:`~radis.misc.arrays.norm`
- :py:func:`~radis.misc.arrays.norm_on`
- :py:func:`~radis.misc.arrays.scale_to`
- :py:func:`~radis.misc.arrays.array_allclose`
- :py:func:`~radis.misc.arrays.nantrapz`
- :py:func:`~radis.misc.arrays.arange_len`
- :py:func:`~radis.misc.arrays.calc_diff`
- :py:func:`~radis.misc.arrays.find_nearest`
- :py:func:`~radis.misc.arrays.find_first`
- :py:func:`~radis.misc.arrays.autoturn`
- :py:func:`~radis.misc.arrays.centered_diff`
- :py:func:`~radis.misc.arrays.evenly_distributed`
- :py:func:`~radis.misc.arrays.anynan`
- :py:func:`~radis.misc.arrays.first_nonnan_index`
- :py:func:`~radis.misc.arrays.last_nonan_index`
- :py:func:`~radis.misc.arrays.is_sorted`
- :py:func:`~radis.misc.arrays.is_sorted_backward`
- :py:func:`~radis.misc.arrays.bining`
- :py:func:`~radis.misc.arrays.count_nans`
- :py:func:`~radis.misc.arrays.logspace`
- :py:func:`~radis.misc.arrays.numpy_add_at`





-------------------------------------------------------------------------------

"""


from math import ceil

import numba
import numpy as np
from numba import bool_, float64, int32, int64
from numpy import hstack
from scipy.interpolate import interp1d

# Normalize


def norm(a, normby=None, how="max"):
    """Normalize a numpy array with its maximum. Or normalize it with another
    vector. Works if array contains nans.

    Parameters
    ----------

    normby: array, or None
        if array, norm with this other array's maximum. If None, normalize with
        its maximum.
    """
    if normby is not None:
        normby = np.abs(normby)
    else:
        normby = np.abs(a)

    if how == "max":
        return a / np.nanmax(normby)
    elif how == "mean":
        return a / np.nanmean(normby)
    else:
        raise ValueError("Unknown normalisation method")


def norm_on(a, w, wmin=None, wmax=None, how="max"):
    """Normalize `a` on a specific range of `w`

    Parameters
    ----------

    a: array
        array
    w: array
        x-axis array

    Other Parameters
    ----------------

    wmin, wmax: float
        crop range
    how: 'mean', 'max'
        how to normalize

    Returns
    -------

    a_norm: array
        normalized array
    """
    imin = np.argmin((w - wmin) ** 2) if wmin else None
    imax = np.argmin((w - wmax) ** 2) if wmax else None
    if imin is not None and imax is not None:
        if imin > imax:
            imin, imax = imax, imin
        elif imin == imax:
            imax += 1
    return norm(a, a[imin:imax], how=how)


def scale_to(a, b, k=1):
    """Scale function a to k*b."""
    return a * k * max(np.abs(b)) / max(np.abs(a))


def array_allclose(a, b, rtol=1e-5, atol=1e-8, equal_nan=True):
    """Returns whether a and b are all close (element wise). If not the same
    size, returns False (instead of crashing like the numpy version). Cf
    numpy.allclose docs for more information.

    Parameters
    ----------

    a, b: arrays
    rtol: float
    atol: float
    equal_nan: bool
        whether to consider NaN's as equal. Contrary to the numpy version this
        one is set to True by default
    """

    if len(a) != len(b):
        return False

    return np.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)


def nantrapz(I, w, dx=1.0, axis=-1):
    """Returns :py:func:`~numpy.trapz` (I, w) discarding nan."""
    b = ~np.isnan(I)
    return np.trapz(I[b], w[b], dx=dx, axis=axis)


# %%
# ==============================================================================
# Numpy Function
# ==============================================================================


def arange_len(wmin, wmax, wstep) -> int:
    """Returns len of a :py:func:`numpy.arange` ``(wmin, max, wstep)`` array,
    accounting for floating point errors

    Note: :py:func:`numpy.arange` is useful to maintain the input ``wstep``.
    If you don't have this requirement, you better use :py:func:`numpy.linspace`
    directly.
    """
    return ceil((wmax - wmin) / wstep)


def calc_diff(t1, v1, t2, v2):
    """Subtract two vectors that may have slightly offset abscissa
    interpolating the correct values.

    Parameters
    ----------

    t1, v1: array_like
        first vector and its abscissa
    t2, v2: array_like
        second vector and its abscissa


    Returns
    -------

    tdiff, vdiff: array_like
        subtracted vector and its abscissa
    """

    t1, v1, t2, v2 = list(map(np.array, (t1, v1, t2, v2)))

    # Deal with inversely sorted case
    if t1[-1] < t1[0]:
        t1, v1 = t1[::-1], v1[::-1]
    if t2[-1] < t2[0]:
        t2, v2 = t2[::-1], v2[::-1]

    # Get the overlapping range
    b = np.logical_and(t2 > t1[0], t2 < t1[-1])

    tdiff = t2[b]
    v2 = v2[b]

    # Interpolate the correct values
    f = interp1d(t1, v1)
    v1 = f(tdiff)

    # Finally, substract:
    vdiff = v1 - v2

    return tdiff, vdiff


def find_nearest(array, searched, return_bool=False):
    """Return the closest elements in array for each element in 'searched'
    array. In case of multiple elements in `array` having equal difference with
    `searched` element, one with least index is returned. Also returns a
    boolean array with indices of elements occurring in output list set to true.

    Examples
    --------

    ::

        from numpy import array
        find_nearest(array([1,2,3,4]), array([2.1,2]), True)

        >>> (array([2, 2]), array([False, True, False, False]))

        find_nearest(np.array([1,2,3,4]), np.array([2.6,2]), True)

        >>> (array([3, 2]), array([False,  True,  True, False]))

        find_nearest(np.array([1, 3]), np.array([2]))

        >>> array([1])

        find_nearest(np.array([3, 1]), np.array([2]))

        >>> array([3])
    """
    if len(array) == 0:
        raise ValueError("Array to be searched cannot be empty")

    b = np.zeros_like(array, dtype=bool)

    def find_nearest(array, value):
        idx = (np.abs(array - value)).argmin()
        return idx, array[idx]

    nearest_els = []
    try:
        for s in searched:
            idx, el = find_nearest(array, s)
            b[idx] = True
            nearest_els.append(el)
    except:
        idx, el = find_nearest(array, searched)
        b[idx] = True
        nearest_els.append(el)

    if return_bool:
        out = nearest_els, b
    else:
        out = nearest_els

    return out


def find_first(arr, threshold):
    """Return the index of the first element of the array arr whose value is
    more than the threshold."""

    return np.argmax(arr > threshold)


def autoturn(data, key=-1):
    """Turns array data. key value:

    - ``0`` don't transpose
    - ``1`` : transpose
    - ``-1`` : auto : make sure the vectors are along the longest dimension
    """

    if key == 0:
        return data
    elif key == 1:
        return data.transpose()
    elif key == -1:
        if data.shape[0] == max(data.shape):
            # Probably columns :
            return data.transpose()
        else:
            return data


def centered_diff(w):
    """Return w[i+1]-w[i-1]/2, same size as w.

    Similar to :py:func:`numpy.diff`, but does not change the array
    size.
    """
    dw = np.diff(w)
    return (hstack((dw, dw[-1])) + hstack((dw[0], dw))) / 2


def evenly_distributed(w, atolerance=1e-5):
    """Make sure array `w` is evenly distributed.

    Parameters
    ----------
    w : numpy array
        array to test
    atolerance: float
        absolute tolerance

    Returns
    -------
    out: bool
        ``True`` or ``False`` if ``w`` is evenly distributed.

    See Also
    --------
    :py:func:`~radis.misc.arrays.evenly_distributed_fast`
    """
    mean_step = np.diff(w).mean()
    return np.allclose(np.diff(w), mean_step, atol=atolerance)


# TODO : rewrite so it works in one loop only ; Jit it


def evenly_distributed_fast(w, rtolerance=1e-5):
    """Make sure array `w` is evenly distributed, by looking only at the first
    and last steps. i.e::

        w[-1] - w[-2] == w[1] - w[0]

    Parameters
    ----------
    w : numpy array
        array to test
    rtolerance: float
        relative tolerance

    Returns
    -------
    out: bool
        ``True`` or ``False`` if ``w`` is evenly distributed.

    See Also
    --------
    :py:func:`~radis.misc.arrays.evenly_distributed`
    """

    return np.isclose(w[-1] - w[-2], w[1] - w[0], rtol=rtolerance)


def anynan(a):
    """Returns whether ``a`` has at least one :py:attr:`~numpy.nan`

    Fastest implementation for arrays with >10^4 elements
    https://stackoverflow.com/a/45011547/5622825
    """
    return np.isnan(np.dot(a, a))


def anynan_vaex(a):
    return not (a.countnan() == 0)


@numba.njit
def first_nonnan_index(a):
    """Returns index of first non-nan value in ``a``

    Returns None is all values are :py:attr:`~numpy.nan`

    See Also
    --------

    :func:`~radis.misc.arrays.last_nonnan_index`
    """
    for i in range(a.size):
        if not np.isnan(a[i]):
            return i
    return None


@numba.njit
def last_nonnan_index(a):
    """Returns index of first non-nan value in ``a``

    Returns None is all values are :py:attr:`~numpy.nan`

    See Also
    --------

    :func:`~radis.misc.arrays.first_nonnan_index`
    """
    for i in range(a.size - 1, 0, -1):
        if not np.isnan(a[i]):
            return i
    return None


@numba.njit
def is_sorted(a):
    """Returns whether ``a`` is sorted in ascending order.

    From B.M. answer on StackOverflow: https://stackoverflow.com/a/47004533/5622825

    See Also
    --------

    :func:`~radis.misc.arrays.is_sorted_backward`
    """
    for i in range(a.size - 1):
        if a[i + 1] < a[i]:
            return False
    return True


@numba.njit
def is_sorted_backward(a):
    """Returns whether ``a`` is sorted in descending order.

    See Also
    --------

    :func:`~radis.misc.arrays.is_sorted`
    """
    for i in range(a.size - 1):
        if a[i + 1] > a[i]:
            return False
    return True


def bining(I, ymin=None, ymax=None, axis=1):
    """Averages a I multi-dimensional array (typically an image) along the y
    axis bining(I) corresponds to I.mean(axis=1) NaN are not taken into
    account.

    Parameters
    ----------

    I: numpy array
        intensity
    ymin: int [0-I.shape[1]]
        If None, 0 is used. Default ``None``.
    ymax: int [0-I.shape[1]]
        If None, I.shape[1] is used. Default ``None``.
    axis: int
        Default 1
    """
    I = np.array(I)  # convert to array in case it's a Pandas dataframe for instance
    if ymin is None:
        ymin = 0
    if ymax is None:
        ymax = I.shape[axis]
    if ymin < 0:
        print("Warning in bining. ymin ({0}) < 0".format(ymin))
    if ymax > I.shape[axis]:
        print(
            "Warning in bining. ymax ({0}) > yaxis length ({1})".format(
                ymax, I.shape[axis]
            )
        )
    return np.nanmean(I[:, ymin:ymax], axis=axis)


def count_nans(a):
    """NaN are good but only in India."""

    return np.isnan(a).sum()


def logspace(xmin, xmax, npoints):
    """Returns points from xmin to xmax regularly distributed on a logarithm
    space.

    Numpy's :py:func:`numpy.logspace` does the same from 10**xmin to 10**xmax
    """

    return np.logspace(np.log10(xmin), np.log10(xmax), npoints)


def numpy_add_at(LDM, k, l, m, I):
    """Add the linestrengths on the LDM grid.

    Uses the numpy implementation of :py:func:`~numpy.add.at`, which
    add arguments element-wise.

    Parameters
    ----------
    LDM : ndarray
        LDM grid
    k, l, m : array
        index
    I : array
        intensity to add

    Returns
    -------
    add: ndarray
        linestrengths distributed over the LDM grid

    Notes
    -----
    Cython version implemented in https://github.com/radis/radis/pull/234

    See Also
    --------
    :py:func:`numpy.add.at`

    """
    # print('Numpy add.at()')
    return np.add.at(LDM, (k, l, m), I)


# ## Cython functions:
# try:
#     import radis_cython_extensions as rcx
# except (ModuleNotFoundError, ValueError):
#     add_at = numpy_add_at
#     """Wrapper to :py:func`radis.misc.arrays.numpy_add_at` or to the Radis
#     Cython version  (here, the numpy version was loaded)"""
#     #  or use radis.misc.utils.NotInstalled() ?
# # EP: Also got (after switching back to former Cython version) :
# #     File "radis\cython\radis_cython_extensions.pyx", line 1, in init radis_cython_extensions
# #     ValueError: numpy.ndarray size changed, may indicate binary incompatibility. Expected 88 from C header, got 80 from PyObject
# # we may consider escaping this.
# else:
#     add_at = rcx.add_at


# @numba.njit
# def non_zero_values_around2(a, n):
#     """return a boolean array of same size as ``a`` where each position ``i``
#     is ``True`` if there are non-zero points less than ``n`` index position
#     away from ``a[i]``, and ``False`` if all points in ``a`` are 0 ``n``  index
#     position away from ``a[i]``
#     """
#     # # Cleaner version, but about 2x slower on real-life test cases:
#     # #
#     # %timeit non_zero_values_around(LDM[:, l, m], truncation)
#     # 2.92 ms ± 99.3 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)

#     # %timeit non_zero_values_around2(LDM[:, l, m], truncation)
#     # 5.11 ms ± 110 µs per loop (mean ± std. dev. of 7 runs, 100 loops each)
#     # --------------

#     # non_zero_values_around2()
#     b = np.zeros(len(a)+2*n+1,  dtype=np.bool_)
#     for pos in np.nonzero(a)[0]:
#         b[pos:pos+2*n+1] = True
#     return b[n:len(a)+n]


@numba.njit(
    bool_[:](float64[:], int64),
    cache=True,
)
def non_zero_values_around(a, n):
    """return a boolean array of same size as ``a`` where each position ``i``
    is ``True`` if there are non-zero points less than ``n`` index position
    away from ``a[i]``, and ``False`` if all points in ``a`` are 0 ``n``  index
    position away from ``a[i]``
    """

    # build the list
    L = []
    pos = -1  # found position of non-zero point
    start = -1  # start position of subrange

    for i, ai in enumerate(a):
        if ai != 0:
            pos = i
            if start == -1:
                if i == 0:
                    start = 0
                else:
                    start = i - n
        else:
            if start == -1:
                continue
            elif pos == i - n - 1:
                L.append((start, i - 1))
                start = -1
                pos = -1

    if start != -1:
        L.append((start, len(a) - 1))

    # turn it into a boolean array
    b = np.zeros(len(a), dtype=np.bool_)
    for start, end in L:
        b[max(0, start) : end + 1] = 1

    return b


# @numba.njit(
#     numba.types.List(numba.types.containers.UniTuple(int64, 2))(int32[:], int64),
#     cache=True,
# )
# def non_zero_ranges_around(a, n):
#     """return a list of coordinates corresponding to non-zero values
#     where each position ``i``
#     is ``True`` if there are non-zero points less than ``n`` index position
#     away from ``a[i]``, and ``False`` if all points in ``a`` are 0 ``n``  index
#     position away from ``a[i]``

#     Parameters
#     ----------
#     a : array of int, line centers
#     n : int or array of int, width
#     """

#     # build the list
#     L = []
#     pos = -1  # found position of non-zero point
#     start = -1  # start position of subrange

#     for i, ai in enumerate(a):
#         if ai != 0:
#             pos = i
#             if start == -1:
#                 if i == 0:
#                     start = 0
#                 else:
#                     start = i - n
#         else:
#             if start == -1:
#                 continue
#             elif pos == i - n - 1:
#                 L.append((start, i - 1))
#                 start = -1
#                 pos = -1

#     if start != -1:
#         L.append((start, len(a) - 1))

#     return L


# @numba.njit(
#     bool_[:](int32[:], int64),
#     cache=True,
# )
# def non_zero_values_around(a, n):
#     """return a boolean array of same size as ``a`` where each position ``i``
#     is ``True`` if there are non-zero points less than ``n`` index position
#     away from ``a[i]``, and ``False`` if all points in ``a`` are 0 ``n``  index
#     position away from ``a[i]``
#     """
#     L = non_zero_ranges_around(a, n)

#     # turn it into a boolean array
#     b = np.zeros(len(a), dtype=np.bool_)
#     for start, end in L:
#         b[max(0, start) : end + 1] = 1

#     return b


@numba.njit(
    (int64[:, :])(bool_[:]),
    cache=True,
)
def non_zero_ranges_in_array(b):
    """return a list of coordinates corresponding to non-zero values
    in boolean array ``b``

    Parameters
    ----------
    b : boolean array

    Examples
    --------
    ::

        b = np.array([0,0,1,1,0,1,0,1], dtype=bool)
        non_zero_ranges_in_array(b)
        >> ([2, 5, 7],
            [4, 6, 8])
    """

    # build the list
    L = []
    start = -1  # start position of subrange

    for i, bi in enumerate(b):
        if bi:
            if start == -1:
                start = i
        else:
            if start == -1:
                continue
            else:
                L.append((start, i))
                start = -1

    if start != -1:
        L.append((start, len(b)))

    return np.array(L)


@numba.njit(
    (bool_[:])(int64[:, :], int64),
    cache=True,
)
def boolean_array_from_ranges(ranges, n):
    """return a boolean array of length ``n`` where (``L[i][0]``, ``L[i][1]``)
    give the ranges set to ``True``

    Examples
    --------
    ::

        L = np.array([[2,4], [5,6], [7,8]])
        boolean_array_from_coordinates(*L.T, 8)
        >>> np.array([0, 0, 1, 1, 0, 1, 0, 1], dtype=bool)

    """
    # build the list
    b = np.zeros(n, dtype=bool_)
    for i in range(len(ranges)):
        b[ranges[i][0] : ranges[i][1]] = True

    return b


@numba.njit(
    numba.types.Tuple((int64[:, :], float64[:]))(
        int32[:], float64[:], float64[:], float64[:], int64, int64
    ),
    cache=True,
)
def sparse_add_at(ki0, Iv0, Iv1, weight, max_range, truncation_pts):
    """Returns (start, stop) of non-zero ranges, and intensity ``I`` (length
    ``truncation_pts``) for all line intensities ``Iv0`` and ``Iv1`` at position ``ki0``
    and ``ki0+1`` with weights ``weight``

    .. warning::
        ``ki0`` must be sorted.
    """
    assert is_sorted(ki0)
    I = np.zeros(max_range)
    L = [
        (0, 0) for k in range(0)
    ]  # empty list; helps Numba infer type. See https://numba.pydata.org/numba-doc/0.41.0/user/troubleshoot.html#my-code-has-an-untyped-list-problem
    n = truncation_pts

    # initiate start
    pos = ki0[0]
    start = max(pos - n, 0)
    end = max_range  # will be rewritten in the first call of the loop

    # Loop over all lines :
    for i in range(len(ki0)):
        pos = ki0[i]

        # Adjust range:
        if pos > end:
            # close the previous range:
            L.append((start, end))
            # reopen new range as `[pos-n, pos+n]`:
            start = max(pos - n, 0)
            end = min(
                pos + 2 + n, max_range
            )  # TODO:  +2 because line is added on ki0 and ki0+1
        else:
            # we're in the middle of a range:
            # keep the start, move the end
            end = min(pos + 2 + n, max_range)

        # Sum intensity  (equivalent of "add-at")
        I[pos] += Iv0[i] * weight[i]
        I[pos + 1] += Iv1[i] * weight[i]
    # final close:
    if len(L) == 0 or L[-1] != (start, end):
        L.append((start, end))

    return np.array(L), I


if __name__ == "__main__":
    import pytest

    pytest.main(["../test/misc/test_arrays.py", "-s"])  # -s for showing console output



# -*- coding: utf-8 -*-
"""
Summary
-----------------------

GEISA database parser

-----------------------


"""


# import re
import time
from collections import OrderedDict
from os.path import exists, getmtime

import numpy as np

import radis

try:
    from .cache_files import load_h5_cache_file, save_to_hdf
    from .tools import drop_object_format_columns, parse_hitran_file
except ImportError:
    if __name__ == "__main__":  # running from this file, as a script
        from radis.api.cache_files import load_h5_cache_file, save_to_hdf
        from radis.io.tools import drop_object_format_columns, parse_hitran_file
    else:
        raise
# from typing import Union
from radis.api.dbmanager import DatabaseManager
from radis.api.hdf5 import DataFileManager

# %% Parsing functions

# General case : GEISA 2020
# Provided by Thibault Delahaye

# BE ADVISED #1: the columns "int" and "ierrB" should be in format FLOAT instead of str as below.
# However, because of Fortran's double-precision floating-point values used in these 2 columns,
# which numpy cannot typecast to float. Thus, we set them as string here and will typecast them
# to float again later in geisa_utils.

# BE ADVISED #2: the columns "idH" for HITRAN molecular number and "isoH" for HITRAN isotope number
# have been changed to "id" and "iso" respectively. Meanwhile, original columns "id" and "iso" have
# also been changed to "idG" and "isoG" to avoid duplication. This is to make sure the parsing process
# and spectrum calculation using HAPI partition function can operate normally.

# fmt: off
columns_GEISA = OrderedDict(
    [
        # name  # format  # type  # description  # unit
        ("wav", ("a12", float, "vacuum wavenumber", "cm-1")),
        ("int", ("a11", str, "intensity at 296K", "cm-1/(molecule/cm-2)")),
        ("airbrd", ("a6", float, "air-broadened half-width at 296K", "cm-1.atm-1")),
        ("El", ("a10", float, "lower-state energy", "cm-1")),
        ("globu", ("a25", str, "electronic and vibrational global upper quanta", "")),
        ("globl", ("a25", str, "electronic and vibrational global lower quanta", "")),
        ("locu", ("a15", str, "electronic and vibrational local upper quanta", "")),
        ("locl", ("a15", str, "electronic and vibrational local lower quanta", "")),
        ("Tdpair", ("a4", float, "temperature-dependance exponent for Gamma air", "")),
        ("isoG", ("a3", int, "GEISA isotope number", "")),
        ("mol", ("a3", int, "GEISA molecular number", "")),
        ("idG", ("a3", str, "Internal GEISA code for the data identification", "")),
        ("id", ("a2", int, "Hitran molecular number", "")),
        ("iso", ("a1", int, "Hitran isotope number", "")),
        ("A", ("a10", float, "Einstein A coefficient", "s-1")),
        ("selbrd", ("a7", float, "self-broadened half-width at 296K", "cm-1.atm-1")),
        ("Pshft", ("a9", float, "air pressure-induced line shift at 296K", "cm-1.atm-1")),
        ("Tdppair", ("a6", float, "temperature-dependance exponent for air pressure-induced line shift", "")),
        ("ierrA", ("a10", float, "estimated accuracy on the line position", "cm-1")),
        ("ierrB", ("a11", str, "estimated accuracy on the intensity of the line", "cm-1/(molecule/cm-2)")),
        ("ierrC", ("a6", float, "estimated accuracy on the air collision halfwidth", "cm-1.atm-1")),
        ("ierrF", ("a4", float, "estimated accuracy on the temperature dependence coefficient of the air-broadening halfwidth", "")),
        ("ierrO", ("a9", float, "estimated accuracy on the air pressure shift of the line transition at 296K", "cm-1.atm-1")),
        ("ierrR", ("a6", float, "estimated accuracy on the temperature dependence coefficient of the air pressure shift", "")),
        ("ierrN", ("a7", float, "estimated accuracy on the self-broadened at 296K", "cm-1.atm-1")),
        ("Tdpself", ("a4", float, "temperature-dependance exponent for self-broadening halfwidth", "")),
        ("ierrS", ("a4", float, "estimated accuracy on the temperature dependence coefficient of the self-broadening halfwidth", "")),
        ("Pshfts", ("a8", float, "self pressure-induced line shift at 296K", "cm-1.atm-1")),
        ("ierrT", ("a8", float, "estimated accuracy on the self-pressure shift of the line transition at 296K", "cm-1.atm-1")),
        ("Tdpnself", ("a4", float, "temperature-dependance exponent for self pressure-induced line shift", "")),
        ("ierrU", ("a4", float, "estimated accuracy on the temperature dependence coefficient of the self pressure shift", "")),
    ]
)
""" OrderedDict: parsing order of GEISA2020 format """

# The columns required for equilibrium calculations only
equilibrium_columns = OrderedDict(
    [
        # name  # format  # type  # description  # unit
        ("wav", ("a12", float, "vacuum wavenumber", "cm-1")),
        ("int", ("a11", float, "intensity at 296K", "cm-1/(molecule/cm-2)")),
        ("airbrd", ("a6", float, "air-broadened half-width at 296K", "cm-1.atm-1")),
        ("iso", ("a3", int, "GEISA isotope number", "")),
        ("mol", ("a3", int, "GEISA molecular number", "")),
        ("id", ("a3", str, "Internal GEISA code for the data identification", "")),
        ("selbrd", ("a7", float, "self-broadened half-width at 296K", "cm-1.atm-1")),
        ("Pshft", ("a9", float, "air pressure-induced line shift at 296K", "cm-1.atm-1")),
        ("Pshfts", ("a8", float, "self pressure-induced line shift at 296K", "cm-1.atm-1")),
    ]
)
# fmt: on

GEISA_MOLECULES_Nlines = {
    "H2O": 362222,
    "CO2": 532533,
    "O3": 389378,
    "N2O": 50633,
    "CO": 14985,
    "CH4": 240858,
    "O2": 6428,
    "NO": 105079,
    "SO2": 68728,
    "NO2": 113880,
    "NH3": 29082,
    "PH3": 34542,
    "HNO3": 691161,
    "OH": 42866,
    "HF": 20010,
    "HCL": 35985,
    "HBR": 8980,
    "HI": 4751,
    "CLO": 7230,
    "OCS": 33809,
    "H2CO": 37050,
    "C2H6": 28439,
    "CH3D": 49237,
    "C2H2": 11340,
    "C2H4": 18378,
    "GEH4": 32372,
    "HCN": 81889,
    "C3H8": 8983,
    "C2N2": 2577,
    "C4H2": 119480,
    "HC3N": 179347,
    "HOCL": 17862,
    "N2": 120,
    "CH3CL": 18344,
    "H2O2": 126983,
    "H2S": 20788,
    "HCOOH": 62684,
    "COF2": 70904,
    "SF6": 92398,
    "C3H4": 19001,
    "HO2": 38804,
    "CLONO2": 356899,
    "CH3BR": 36911,
    "CH3OH": 19897,
    "NO+": 1206,
    "HNC": 5619,
    "C6H6": 9797,
    "C2HD": 15512,
    "CF4": 60033,
    "CH3CN": 17172,
    "HDO": 63641,
    "SO3": 10881,
    "HONO": 26041,
    "COFCL": 215639,
    "CH3I": 70291,
    "CH3F": 1499,
    "RUO4": 30205,
    "H2C3H2": 31686,
}

GEISA_MOLECULES = list(GEISA_MOLECULES_Nlines.keys())


def gei2df(
    fname,
    cache=True,
    load_columns=None,
    verbose=True,
    drop_non_numeric=True,
    load_wavenum_min=None,
    load_wavenum_max=None,
    engine="pytables",
):
    """Convert a GEISA [1]_ file to a Pandas dataframe.
    Parameters
    ----------
    fname: str
        GEISA file name.
    cache: boolean, or 'regen'
        if ``True``, a pandas-readable HDF5 file is generated on first access,
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot (but changes in the database are not
        taken into account). If ``False``, no database is used. If 'regen', temp
        file are reconstructed. Default ``True``.
    load_columns: list
        columns to load. If ``None``, loads everything.
        .. note::
            this is only relevant when loading from a cache file. To generate
            the cache file, all columns are loaded anyway.
    Other Parameters
    ----------------
    drop_non_numeric: boolean
        if ``True``, non numeric columns are dropped. This improves performances,
        but make sure all the columns you need are converted to numeric formats
        before hand. Default ``True``. Note that if a cache file is loaded it
        will be left untouched.
    load_wavenum_min, load_wavenum_max: float
        if not ``'None'``, only load the cached file if it contains data for
        wavenumbers above/below the specified value. See :py:func`~radis.api.cache_files.load_h5_cache_file`.
        Default ``'None'``.
    engine: 'pytables', 'vaex'
        format for Hdf5 cache file, `pytables` by default.
    Returns
    -------
    df: pandas Dataframe
        dataframe containing all lines and parameters.
    Notes
    -----
    GEISA Database 2020 release can be downloaded from [2]_
    References
    ----------
    .. [1] `The 2020 edition of the GEISA spectroscopic database, Thibault Delahaye et al., 2021 <https://www.sciencedirect.com/science/article/abs/pii/S0022285221000928>`_
    .. [2] `GEISA Database 2020 release <https://geisa.aeris-data.fr/interactive-access/?db=2020&info=ftp>`_
    See Also
    --------
    :func:`~radis.io.hitran.hit2df`
    :func:`~radis.io.cdsd.cdsd2df`
    """

    # Try to access last modification time of original file
    metadata = {}
    metadata["last_modification"] = time.ctime(getmtime(fname))
    # Make sure the wavenums are legit
    if load_wavenum_min and load_wavenum_max:
        assert load_wavenum_min < load_wavenum_max
    # Verbose
    if verbose >= 2:
        print("Opening file {0}, cache={1})".format(fname, cache))
        print("Last Modification time: {0}".format(metadata["last_modification"]))

    # Attempt to use cache file
    fcache = DataFileManager(engine).cache_file(fname)
    if cache and exists(fcache):
        relevant_if_metadata_above = (
            {"wavenum_max": load_wavenum_min} if load_wavenum_min else {}
        )  # not relevant if wavenum_max of file is < wavenum min required
        relevant_if_metadata_below = (
            {"wavenum_min": load_wavenum_max} if load_wavenum_max else {}
        )  # not relevant if wavenum_min of file is > wavenum max required
        df = load_h5_cache_file(
            fcache,
            cache,
            columns=load_columns,
            valid_if_metadata_is=metadata,
            relevant_if_metadata_above=relevant_if_metadata_above,
            relevant_if_metadata_below=relevant_if_metadata_below,
            current_version=radis.__version__,
            last_compatible_version=radis.config["OLDEST_COMPATIBLE_VERSION"],
            verbose=verbose,
            engine=engine,
        )
        if df is not None:
            return df

    # If cache files are not found, commence reading of full file
    df = parse_hitran_file(fname, columns_GEISA)

    # Commence "D to E" conversion on 2nd and 20th columns, by using
    # df.columns.str.replace() which is much faster than Python loop.
    # Finally, typecast them to float.
    df["int"] = df["int"].astype(str).str.replace("D", "E").astype(float)
    df["ierrB"] = df["ierrB"].astype(str).str.replace("D", "E").astype(float)

    # Remove non numerical attributes
    if drop_non_numeric:
        # The function replace_PQR_with_m101 below will be developed
        # in later updates for GEISA's non-equilibrium calculations
        # replace_PQR_with_m101(df)

        # Basically ripping out 4 quanta columns
        df = drop_object_format_columns(df, verbose=verbose)

    # Generate cache file for later use
    if cache:
        new_metadata = {
            # Last modification time of the original file :
            "last_modification": time.ctime(getmtime(fname)),
            "wavenum_min": df.wav.min(),
            "wavenum_max": df.wav.max(),
        }
        if verbose:
            print(
                "Generating cache file {0} with metadata :\n{1}".format(
                    fcache, new_metadata
                )
            )
        try:
            save_to_hdf(
                df,
                fcache,
                metadata=new_metadata,
                version=radis.__version__,
                key="df",
                overwrite=True,
                verbose=verbose,
                engine=engine,
            )
        except PermissionError:
            if verbose:
                print(
                    "An error occurred in cache file generation. Lookup access rights."
                )
            pass

    return df


#%%
def get_last(b):
    """Get non-empty lines of a chunk b, parsing the bytes."""
    element_length = np.vectorize(lambda x: len(x.__str__()))(b)
    non_zero = element_length > element_length[-1]
    threshold = non_zero.argmin() - 1
    assert (non_zero[: threshold + 1] == 1).all()
    assert (non_zero[threshold + 1 :] == 0).all()
    return b[non_zero]


class GEISADatabaseManager(DatabaseManager):
    def __init__(
        self,
        name,
        molecule,
        local_databases,
        engine="default",
        verbose=True,
        chunksize=100000,
        parallel=True,
    ):
        super().__init__(
            name,
            molecule,
            local_databases,
            engine,
            verbose=verbose,
            parallel=parallel,
        )
        self.chunksize = chunksize
        self.downloadable = True
        self.base_url = None
        self.Nlines = None
        self.wmin = None
        self.wmax = None
        self.urlnames = None

    def fetch_urlnames(self):
        r"""requires connexion"""

        if self.urlnames is not None:
            return self.urlnames

        molecule = self.molecule
        geisa_url = "https://aeris-geisa.ipsl.fr/geisa_files/2020/Lines/line_GEISA2020_asc_gs08_v1.0_"

        print(f"Molecule: {molecule}")

        if molecule.upper() in GEISA_MOLECULES:
            url = geisa_url + molecule.lower()
            urlnames = [url]
        else:
            raise KeyError(
                f"Please choose one of GEISA molecules : {GEISA_MOLECULES}. Got '{molecule}'"
            )

        self.urlnames = urlnames

        return urlnames

    def parse_to_local_file(
        self,
        opener,
        urlname,
        local_file,
        pbar_active=True,
        pbar_t0=0,
        pbar_Ntot_estimate_factor=None,
        pbar_Nlines_already=0,
        pbar_last=True,
    ):
        r"""Uncompress ``urlname`` into ``local_file``.
        Also add metadata

        Parameters
        ----------
        opener: an opener with an .open() command
        gfile : file handler. Filename: for info"""

        # molecule = self.molecule
        # Ntotal_lines_expected = GEISA_MOLECULES_Nlines[molecule.upper()]

        writer = self.get_datafile_manager()

        with opener.open(urlname) as gfile:  # locally downloaded file

            gfile  #  so the linter doesn't annoy us. We're not using this file anyway, just unzipping the cache file directly :
            df = gei2df(opener.abspath(urlname), drop_non_numeric=False, cache=False)

            writer.write(local_file, df, append=False)

            self.wmin = df.wav.min()
            self.wmax = df.wav.max()

            Nlines = len(df)

        writer.combine_temp_batch_files(local_file)  # used for vaex mode only

        # Check number of lines is consistent
        # assert Nlines == Ntotal_lines_expected

        # Add metadata
        from radis import __version__

        writer.add_metadata(
            local_file,
            {
                "wavenumber_min": self.wmin,
                "wavenumber_max": self.wmax,
                "download_date": self.get_today(),
                "download_url": urlname,
                "total_lines": Nlines,
                "version": __version__,
            },
        )

        return Nlines

    def register(self):
        r"""register in ~/radis.json"""

        local_files, urlnames = self.get_filenames()
        info = f"GEISA {self.molecule} lines ({self.wmin:.1f}-{self.wmax:.1f} cm-1)"

        dict_entries = {
            "info": info,
            "path": local_files,
            "format": "geisa-radisdb",
            "parfuncfmt": "hapi",
            "wavenumber_min": self.wmin,
            "wavenumber_max": self.wmax,
            "download_date": self.get_today(),
            "download_url": urlnames,
        }

        super().register(dict_entries)


#%%

if __name__ == "__main__":

    import pytest

    print("Testing factory:", pytest.main(["../test/io/test_geisa.py"]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Routine Listing
---------------

This module contains...

a function to convolve with a slit function, that can resample and correct
for slit dilation:

- :func:`~radis.tools.slit.convolve_with_slit`

This function is called directly by the Spectrum method
:py:meth:`~radis.spectrum.spectrum.Spectrum.apply_slit`

predefined slit functions generators (triangular, gaussian, etc... see :data:`~radis.tools.slit.SLIT_SHAPES`)
and experimental slit function importer:

- :func:`~radis.tools.slit.get_slit_function`
- :func:`~radis.tools.slit.import_experimental_slit`
- :func:`~radis.tools.slit.triangular_slit`
- :func:`~radis.tools.slit.trapezoidal_slit`
- :func:`~radis.tools.slit.gaussian_slit`

functions to manipulate slit functions (plot, get_FWHM, get_effective_FWHM,
recenter_slit, crop_slit):

- :func:`~radis.tools.slit.get_FWHM`
- :func:`~radis.tools.slit.get_effective_FWHM`
- :func:`~radis.tools.slit.plot_slit`
- :func:`~radis.tools.slit.recenter_slit`
- :func:`~radis.tools.slit.crop_slit`


-------------------------------------------------------------------------------



"""


from warnings import warn

import numpy as np
from numpy import exp
from numpy import log as ln
from numpy import sqrt, trapz
from scipy.interpolate import splev, splrep
from scipy.signal import oaconvolve

from radis.misc.arrays import anynan, evenly_distributed, evenly_distributed_fast
from radis.misc.basics import is_float
from radis.misc.debug import printdbg
from radis.misc.signal import resample_even
from radis.misc.warning import SlitDispersionWarning
from radis.phys.convert import cm2nm, dcm2dnm, dnm2dcm, nm2cm
from radis.spectrum.spectrum import cast_waveunit

SLIT_SHAPES = ["triangular", "trapezoidal", "gaussian"]
"""list : list of predefined slit shapes
"""

# %% Get slit function


def get_slit_function(
    slit_function,
    unit="nm",
    norm_by="area",
    shape="triangular",
    center_wavespace=None,
    return_unit="same",
    wstep=None,
    plot=False,
    resfactor=2,
    verbose=True,
    auto_recenter_crop=True,
    *args,
    **kwargs,
):
    """Import or generate slit function in correct wavespace
    Give a file path to import, or a float / tuple to generate arbitrary shapes

    Warning with units: read about unit and return_unit parameters.

    See :meth:`~radis.spectrum.spectrum.Spectrum.apply_slit` and
    :func:`~radis.tools.slit.convolve_with_slit` for more info


    Parameters
    ----------

    slit_function: tuple, or str
        If float:
            generate slit function with FWHM of `slit_function` (in `unit`)
        If .txt file:
            import experimental slit function (in `unit`): format must be 2-columns
            with wavelengths and intensity (doesn't have to be normalized)

    unit: 'nm' or 'cm-1'
        unit of slit_function FWHM, or unit of imported file

    norm_by: ``'area'``, ``'max'``, or ``None``
        how to normalize. ``'area'`` conserves energy. With ``'max'`` the slit is normalized
        at peak so that the maximum is one.

        .. note::

            ``'max'`` changes the unit of the spectral array, e.g. from
            ``'mW/cm2/sr/µm'`` to ``'mW/cm2/sr')``

        ``None`` doesnt normalize. Default ``'area'``

    shape: ``'triangular'``, ``'trapezoidal'``, ``'gaussian'``
        which shape to use when generating a slit. Default ``'triangular'``.
        Should be one of :py:data:`~radis.tools.slit.SLIT_SHAPES`.

    center_wavespace: float, or None
        center of slit when generated (in unit). Not used if slit is imported.

    return_unit: ``'nm'``, ``'cm-1'``, or ``'same'``
        if not ``'same'``, slit is converted to the given wavespace.
        Default ``'same'``.

    wstep: float
        which discretization step to use (in return_unit) when generating a slit
        function. Not used if importing


    Other Parameters
    ----------------

    resfactor: int
        resolution increase when resampling from nm to cm-1, or the other way
        round. Default 2.

    energy_threshold: float
         tolerance fraction. Only used when importing experimental slit as the
         theoretical slit functions are directly generated in spectrum wavespace
         Default 1e-3 (0.1%)

    auto_recenter_crop: bool
        if ``True``, recenter slit and crop zeros on the side when importing
        an experimental slit. Default ``True``.
        See :func:`~radis.tools.slit.recenter_slit`, :func:`~radis.tools.slit.crop_slit`


    Returns
    -------

    wslit, Islit: array
        wslit is in `return_unit` . Islit is normalized according to  `norm_by`


    Examples
    --------

    >>> wslit, Islit = get_slit_function(1, 'nm', shape='triangular',
    center_wavespace=600, wstep=0.01)

    Returns a triangular slit function of FWHM = 1 nm, centered on 600 nm, with
    a step of 0.01 nm

    >>> wslit, Islit = get_slit_function(1, 'nm', shape='triangular',
    center_wavespace=600, return_unit='cm-1', wstep=0.01)

    Returns a triangular slit function expressed in cm-1, with a FWHM = 1 nm
    (converted in equivalent width in cm-1 at 600 nm), centered on 600 nm, with
    a step of 0.01 cm-1 (!)


    Notes
    -----

    In ``norm_by='max'`` mode, slit is normalized by slit max. In RADIS, this is done
    in the spectrum wavespace (to avoid errors that would be caused by interpolating
    the spectrum).

    A problem arise if the spectrum wavespace is different from the slit wavespace:
    typically, slit is in ``'nm'`` but a spectrum calculated by RADIS is stored in
    ``'cm-1'``: in that case, the convoluted spectrum is divided by ``int(Islit*dν)``
    instead of ``int(Islit*dλ)``. The output unit is then ``[radiance]*[spec_unit]``
    instead of ``[radiance]*[slit_unit]``, i.e. typically, ``[mW/cm2/sr/nm]*[cm-1]``
    instead of ``[mW/cm2/sr/nm]*[nm]=[mW/cm2/sr]``

    While this remains true if the units are taken into account, this is not the
    expected behaviour. For instance, Specair users are used to having a FWHM(nm)
    factor between spectra convolved with slit normalized by max and slit normalized
    by area.

    The ``norm_by='max'`` behaviour adds a correction factor
    ```/int(Islit*dλ)/int(Islit*dν)``` to maintain an output spectrum in ``[radiance]*[slit_unit]``

    See Also
    --------

    :meth:`~radis.spectrum.spectrum.Spectrum.apply_slit`,
    :func:`~radis.tools.slit.convolve_with_slit`

    """

    if "waveunit" in kwargs:
        assert return_unit == "same"  # default
        return_unit = kwargs.pop("waveunit")
        warn(DeprecationWarning("waveunit renamed return_unit"))
    if "slit_unit" in kwargs:
        assert unit == "nm"  # default
        unit = kwargs.pop("slit_unit")
        warn(DeprecationWarning("slit_unit renamed unit"))

    energy_threshold = kwargs.pop("energy_threshold", 1e-3)  # type: float
    # tolerance fraction
    # when resampling (only used in experimental slit as the)
    # theoretical slit functions are directly generated in
    # spectrum wavespace

    def check_input_gen():
        if center_wavespace is None:
            raise ValueError(
                "center_wavespace has to be given when generating " + "slit function"
            )
        if wstep is None:
            raise ValueError("wstep has to be given when generating " + "slit function")

    # Cast units
    if return_unit == "same":
        return_unit = unit
    unit = cast_waveunit(unit)
    return_unit = cast_waveunit(return_unit)
    # in norm_by=max mode, used to keep units in [Iunit]*return_unit in [Iunit]*unit
    scale_slit = 1
    # not used in norm_by=area mode

    # Generate Slit
    # -------------
    # First get the slit in return_unit space
    if is_float(
        slit_function
    ):  # Generate slit function (directly in return_unit space)

        check_input_gen()

        # ... first get FWHM in return_unit  (it is in `unit` right now)
        FWHM = slit_function
        if return_unit == "cm-1" and unit == "nm":
            # center_wavespace ~ nm, FWHM ~ nm
            FWHM = dnm2dcm(FWHM, center_wavespace)  # wavelength > wavenumber
            center_wavespace = nm2cm(center_wavespace)
            if norm_by == "max":
                scale_slit = slit_function / FWHM  # [unit/return_unit]
        elif return_unit == "nm" and unit == "cm-1":
            # center_wavespace ~ cm-1, FWHM ~ cm-1
            FWHM = dcm2dnm(FWHM, center_wavespace)  # wavenumber > wavelength
            center_wavespace = cm2nm(center_wavespace)
            if norm_by == "max":
                scale_slit = slit_function / FWHM  # [unit/return_unit]
        else:
            pass  # correct unit already
        # Now FWHM is in 'return_unit'

        # ... now, build it (in our wavespace)
        if __debug__:
            printdbg(
                "get_slit_function: {0} FWHM {1:.2f}{2}, center {3:.2f}{2}, norm_by {4}".format(
                    shape, FWHM, return_unit, center_wavespace, norm_by
                )
            )

        if shape == "triangular":
            wslit, Islit = triangular_slit(
                FWHM,
                wstep,
                center=center_wavespace,
                bplot=plot,
                norm_by=norm_by,
                wunit=return_unit,
                scale=scale_slit,
                *args,
                **kwargs,
            )

        # Insert other slit shapes here
        # ...

        elif shape == "gaussian":
            wslit, Islit = gaussian_slit(
                FWHM,
                wstep,
                center=center_wavespace,
                bplot=plot,
                norm_by=norm_by,
                wunit=return_unit,
                scale=scale_slit,
                *args,
                **kwargs,
            )

        elif shape == "trapezoidal":
            raise TypeError("A (top, base) tuple must be given with a trapezoidal slit")

        else:
            raise TypeError(
                "Slit function ({0}) not in known slit shapes: {1}".format(
                    shape, SLIT_SHAPES
                )
            )

    elif isinstance(slit_function, tuple):

        check_input_gen()

        try:
            top, base = slit_function
        except:
            raise TypeError("Wrong format for slit function: {0}".format(slit_function))
        if shape == "trapezoidal":
            pass
        elif shape == "triangular":  # it's the default
            warn("Triangular slit given with a tuple: we used trapezoidal slit instead")
            shape = "trapezoidal"
        else:
            raise TypeError("A (top, base) tuple must be used with a trapezoidal slit")

        # ... first get FWHM in our wavespace unit
        if return_unit == "cm-1" and unit == "nm":
            # center_wavespace ~ nm, FWHM ~ nm
            top = dnm2dcm(top, center_wavespace)  # wavelength > wavenumber
            base = dnm2dcm(base, center_wavespace)  # wavelength > wavenumber
            center_wavespace = nm2cm(center_wavespace)
            if norm_by == "max":
                scale_slit = sum(slit_function) / (top + base)  # [unit/return_unit]
        elif return_unit == "nm" and unit == "cm-1":
            # center_wavespace ~ cm-1, FWHM ~ cm-1
            top = dcm2dnm(top, center_wavespace)  # wavenumber > wavelength
            base = dcm2dnm(base, center_wavespace)  # wavenumber > wavelength
            center_wavespace = cm2nm(center_wavespace)
            if norm_by == "max":
                scale_slit = sum(slit_function) / (top + base)  # [unit/return_unit]
        else:
            pass  # correct unit already

        FWHM = (top + base) / 2

        # ... now, build it (in our wavespace)
        if __debug__:
            printdbg(
                "get_slit_function: {0}, FWHM {1:.2f}{2}, center {3:.2f}{2}, norm_by {4}".format(
                    shape, FWHM, return_unit, center_wavespace, norm_by
                )
            )

        wslit, Islit = trapezoidal_slit(
            top,
            base,
            wstep,
            center=center_wavespace,
            bplot=plot,
            norm_by=norm_by,
            wunit=return_unit,
            scale=scale_slit,
            *args,
            **kwargs,
        )

    # Or import it from file or numpy input
    # ------------
    elif isinstance(slit_function, str) or isinstance(
        slit_function, np.ndarray
    ):  # import it
        if __debug__:
            printdbg(
                "get_slit_function: {0} in {1}, norm_by {2}, return in {3}".format(
                    slit_function, unit, norm_by, return_unit
                )
            )

        wslit, Islit = import_experimental_slit(
            slit_function,
            norm_by=norm_by,  # norm is done later anyway
            wunit=unit,
            verbose=verbose,
            auto_crop=auto_recenter_crop,
            auto_recenter=auto_recenter_crop,
            bplot=False,  # we will plot after resampling
            *args,
            **kwargs,
        )

        # ... get unit
        # Normalize
        if norm_by == "area":  # normalize by the area
            #        I_slit /= np.trapz(I_slit, x=w_slit)
            Iunit = "1/{0}".format(unit)
        elif norm_by == "max":  # set maximum to 1
            Iunit = ""
        elif norm_by is None:
            Iunit = None
        else:
            raise ValueError(
                "Unknown normalization type: `norm_by` = {0}".format(norm_by)
            )

        # ... check it looks correct
        unq, counts = np.unique(wslit, return_counts=True)
        dup = counts > 1
        if dup.sum() > 0:
            raise ValueError(
                "Not all wavespace points are unique: slit function "
                + "format may be wrong. Duplicates for w={0}".format(unq[dup])
            )

        # ... resample if needed
        if return_unit == "cm-1" and unit == "nm":  # wavelength > wavenumber
            wold, Iold = wslit, Islit
            wslit, Islit = resample_even(
                nm2cm(wslit),
                Islit,
                resfactor=resfactor,
                energy_threshold=energy_threshold,
                print_conservation=True,
            )
            scale_slit = trapz(Iold, wold) / trapz(Islit, wslit)  # [unit/return_unit]
            renormalize = True
        elif return_unit == "nm" and unit == "cm-1":  # wavenumber > wavelength
            wold, Iold = wslit, Islit
            wslit, Islit = resample_even(
                cm2nm(wslit),
                Islit,
                resfactor=resfactor,
                energy_threshold=energy_threshold,
                print_conservation=True,
            )
            scale_slit = trapz(Iold, wold) / trapz(Islit, wslit)  # [unit/return_unit]
            renormalize = True
        else:  # return_unit == unit
            renormalize = False
        # Note: if wstep dont match with quantity it's alright as it gets
        # interpolated in the `convolve_with_slit` function

        # re-Normalize if needed (after changing units)
        if renormalize:
            if __debug__:
                printdbg("get_slit_function: renormalize")
            if norm_by == "area":  # normalize by the area
                Islit /= abs(np.trapz(Islit, x=wslit))
                Iunit = "1/{0}".format(return_unit)
            elif norm_by == "max":  # set maximum to 1
                Islit /= abs(np.max(Islit))
                Islit *= scale_slit
                Iunit = ""
                if scale_slit != 1:
                    Iunit += "x{0}".format(scale_slit)
            elif norm_by is None:
                Iunit = None
            else:
                raise ValueError(
                    "Unknown normalization type: `norm_by` = {0}".format(norm_by)
                )

        if plot:  # (plot after resampling / renormalizing)
            # Plot slit
            plot_slit(wslit, Islit, wunit=return_unit, Iunit=Iunit)

    else:
        raise TypeError(
            "Unexpected type for slit function: {0}".format(type(slit_function))
        )

    # Final shape
    if len(wslit) < 5:
        raise ValueError(
            "Slit should have at least 5 points. Got {0} only. ".format(len(wslit))
            + "Reduce `wstep=`?"
        )

    return wslit, Islit


try:  # Python >3.6 only
    get_slit_function.__annotations__["shape"] = SLIT_SHAPES
except AttributeError:
    pass  # old Python version

# %% Convolve with slit function


def convolve_with_slit(
    w,
    I,
    w_slit,
    I_slit,
    mode="valid",
    k=1,
    bplot=False,
    verbose=True,
    assert_evenly_spaced=True,
    wunit="",
    waveunit=None,
):
    """Convolves spectrum (w,I) with instrumental slit function (w_slit, I_slit)

        Returns a convolved spectrum on a valid range.

        Check Notes section below for more details about its implementation.

        This function is called directly from a :py:class:`~radis.spectrum.spectrum.Spectrum`
        object with the :py:meth:`~radis.spectrum.spectrum.Spectrum.apply_slit` method.


    .. warning::

        By default, convoluted spectra are thinner than non convoluted spectra,
        as spectra are cropped to remove boundary effects. See ``mode=`` to change
        this behaviour.

    Parameters
    ----------

    w, I: array
        theoretical spectrum (wavespace unit (nm / cm-1))

    w_slit, I_slit: array
        instrumental slit function  (wavespace unit (nm / cm-1))
        It is recommended to truncate the input slit function to its minimum spectral
        extension (see Notes).

        .. warning::

            Both wavespaces have to be the same!

    mode: ``'valid'``, ``'same'``

        ``'same'``:
        returns output of same length as initial spectra,
        but boundary effects are still visible.

        ``'valid'``:
        returns output of length len(spectra) - len(slit) + 1, assuming len(I_slit) < len(I),
        for which lines outside of the calculated range have no impact.

        Default ``'valid'``.

    Other Parameters
    ----------------
    k: int
        order of spline interpolation. 3: cubic, 1: linear. Default 1.
    bplot: boolean
        if ``True``, plots the convolve slit function (for debugging)
    verbose: boolean
        more blabla
    assert_evenly_spaced: boolean, or ``'resample'``
        for the convolution to be accurate, ``w`` should be evenly spaced. If
        ``assert_evenly_spaced=True``, then we check this is the case, and raise
        an error if arrays is not evenly spaced. If ``'resample'``, then we resample
        ``w`` and ``I`` if needed. Recommended, but it takes some time.
    wunit: ``'nm'``, ``'cm-1'``
        just for printing messages. However, ``w`` and ``w_slit`` should be in the
        same wavespace.

    Returns
    -------

    w_conv, I_conv: arrays
        convoluted quantity. w_conv is defined on a valid range
        (sides are cut) if ``mode='valid'``

    Notes
    -----
    The discrete convolution operation is defined as

    .. math:: (I * I_{slit})[n] = \\sum_{m = -\\infty}^{\\infty} I[m] I_{slit}[n - m]

    This numerical discrete convolution doesn't take care of respecting the physical distances.
    Therefore we have to make sure that the ``I`` and ``I_slit`` are expressed using the same
    spectral spacing (e.g. that ``w`` and ``w`_slit`` have the same step in nm between
    two adjacent points). The minimal step between values of ``w`` is used to determine
    this reference spacing.

    If the slit function covers a spectral interval much larger than its
    Full Width a Half Maximum (FWHM), it will potentially add non-zero elements in the sum
    and mix spectral features that are much further away from each other than the FWHM, as
    it would have been expected.
    For an experimental slit function, this also means introducing more noise and artificial
    broadening. It is thus recommended to truncate the input slit function to its minimum
    useful spectral range.

    For instance, if you have an instrumental slit function that looks like so, we recommend
    to cut it where the vertical lines are drawn.

    ::

        I_slit  ^
                |     |   _   |
                |     |  / \\  |
                |     | /   \\ |
               0|_____|/     \\|________
              --|-----         --------->
                    cut       cut        w_slit

    When we apply a slit function, we should consider the contributions of spectral features outside
    of the initial spectral range (i.e. ``w[0]`` - FWHM, ``w[-1]`` + FWHM) that would matter to reproduce
    the entire spectrum. Those points are probably not taken into account during the simulation and
    it could be interesting to extend the spectral range of ``w`` when possible to gauge this contribution.

    Since, we don't have access to this extra points, we truncated the output spectrum to the spectral
    range we can trust (no missing information). The ``valid`` mode operates this truncation by cutting
    the edge of the spectrum ever a spectral width corresponding to the spectral width of the provided
    slit function. This is meant to remove incorrect points on the edge. We recommend using this mode.
    You can however access those removed points using the ``same`` mode (at your own risk).

    Implementation of :func:`~radis.tools.slit.convolve_with_slit` is done in 5 steps:

    - Check input slit function: compare input slit spectral range to spectrum spectral range
      if ranges are comparable, check FWHM and send a warning.
    - Interpolate the slit function on the spectrum grid, resample it if not
      evenly spaced (in order to match physical distances)
    - Check slit aspect, plot slit if asked for
    - Convolve!
    - Remove boundary effects

    See Also
    --------

    :func:`~radis.tools.slit.get_slit_function`,
    :meth:`~radis.spectrum.spectrum.Spectrum.apply_slit`

    """

    # 1. Check input
    # --------------

    # Deprecated input:
    if waveunit is not None:
        warn(
            "`waveunit=` parameter in convolve_with_slit is now named `wunit=`",
            DeprecationWarning,
        )
        wunit = waveunit

    # Assert slit function is thin enough

    w_range = abs(w[-1] - w[0])
    w_slit_range = abs(w_slit[-1] - w_slit[0])
    slit_FWHM = get_FWHM(w_slit, I_slit)

    if w_slit_range > 10 * slit_FWHM:
        warn(
            f"FWHM >> spectral range!\n"
            + f"Slit Function is provided with a spectral range {w_slit_range:.1f} nm that is much "
            + f"wider than its estimated FWHM {slit_FWHM:.1f}. We recommend truncating the input "
            + f"slit function. See radis documentation about convolve_with_slit for more information."
        )

    if w_range < w_slit_range:
        if mode == "valid":
            raise AssertionError(
                f"Slit function is provided with a spectral range ({w_slit_range:.1f} {waveunit}) larger than "
                + f"the spectral range of the spectrum ({w_range:.1f} {waveunit}) : the output spectrum will therefore be empty "
                + f"as boundary effects of the convolution are automatically discarded. If you still want to apply the slit "
                + f"despite potential boundary effects, set `mode='same'`. However, we recommend truncating the input slit function if possible, or compute the spectrum on a larger spectral range. "
                + f"See radis documentation about convolve_with_slit for more information."
            )
        else:
            warn(
                f"Slit spectral range > Spectrum spectral range!\n"
                + f"Slit function is provided with a spectral range ({w_slit_range:.1f} nm) larger than "
                + f"the spectral range of the spectrum ({w_range:.1f} nm). We recommend truncating the input "
                + f"slit function. See radis documentation about convolve_with_slit for more information."
            )

    # 2. Interpolate the slit function on the spectrum grid, resample it if not
    #    evenly spaced
    # --------------

    # check first & last spacing, always :
    is_evenly_spaced = evenly_distributed_fast(w, rtolerance=1e-3)
    wstep = w[1] - w[0]

    # ... check with details
    if is_evenly_spaced and assert_evenly_spaced:
        wstep = abs(np.diff(w)).min()  # spectrum wavelength spacing
        is_evenly_spaced = evenly_distributed(w, atolerance=wstep * 1e-3)

    if not is_evenly_spaced:
        if assert_evenly_spaced == "resample":
            w, I = resample_even(w, I, resfactor=2, print_conservation=True)
            wstep = abs(np.diff(w)).min()  # new spectrum wavelength spacing

        raise ValueError(
            "Spectrum is not evenly spaced. Cannot apply slit function. Please resample with `convolve_with_slit(..., convolve_with_slit='resample')` ; or if using a Spectrum object `s`, with `s.resample_even()`"
        )

    # ... Check that the slit is not reversed (interpolation requires objects are sorted)
    reverse = w_slit[-1] < w_slit[0]
    if reverse:
        w_slit = w_slit[::-1]
        I_slit = I_slit[::-1]

    if not np.allclose(np.diff(w_slit), wstep):
        if verbose >= 2:
            # numerical errors can be produced
            print("interpolating slit function over spectrum grid")
        try:
            tck = splrep(w_slit, I_slit, k=k)
        except ValueError:
            # Probably error on input data. Print it before crashing.
            print("\nValueError - Input data below:")
            print("-" * 5)
            print(w_slit)
            print(I_slit)
            print("Check figure")
            plot_slit(w_slit, I_slit, wunit=wunit)
            raise

        # can be a bug here if wstep has the wrong sign.
        w_slit_interp = np.arange(w_slit[0], w_slit[-1] + wstep, wstep)
        # more test needed.
        I_slit_interp = splev(w_slit_interp, tck)
    else:
        w_slit_interp = w_slit
        I_slit_interp = I_slit

    # 3. Check aspect
    # -------------

    # Check no nan
    if anynan(I_slit):
        raise ValueError("Slit has nan value")

    # check slit is positive
    if not (I_slit >= 0).all():
        plot_slit(w_slit, I_slit, wunit="")
        raise ValueError("Slit is partially negative. Check Figure")

    # Plot slit if asked for
    if bplot:
        plot_slit(w_slit, I_slit, wunit="")

    # 4. Convolve!
    # --------------

    # We actually do not use mode valid in np.convolve,
    # instead we use mode=same and remove the same boundaries from I and W in remove_boundary()
    I_conv = oaconvolve(I, I_slit_interp, mode="same") * wstep

    # 5. Remove boundary effects
    # --------------

    w_conv, I_conv = remove_boundary(
        w, I_conv, mode, len_I=len(I), len_I_slit_interp=len(I_slit_interp)
    )

    # reverse back if needed
    # Todo: add test case for that
    if reverse:
        w_slit = w_slit[::-1]
        I_slit = I_slit[::-1]

    return w_conv, I_conv


# %% Slit function methods


def get_FWHM(w, I, return_index=False):
    """Calculate full width half maximum (FWHM) by comparing amplitudes


    Parameters
    ----------

    w, I: arrays

    return_index: boolean
        if True, returns indexes for half width boundaries. Default ``False``

    Returns
    -------

    FWHM: float
        full width at half maximum

    [xmin, xmax]: int


    See Also
    --------

    :py:func:`~radis.tools.slit.get_effective_FWHM`,
    :py:func:`~radis.tools.slit.offset_dilate_slit_function`,
    :py:func:`~radis.tools.slit.normalize_slit`,
    :py:func:`~radis.tools.slit.remove_boundary`,
    :py:func:`~radis.tools.slit.plot_slit`,
    :py:func:`~radis.tools.slit.recenter_slit`,
    :py:func:`~radis.tools.slit.crop_slit`

    """
    # TODO: Linearly interpolate at the boundary? insignificant for large number of points

    half = I.max() / 2
    I_offset = I - half
    upper = np.argwhere(I_offset >= 0)

    xmin = upper.min()
    xmax = upper.max()

    def get_zero(x1, y1, x2, y2):
        r"""
        Linear interpolation on data to get zero
        Return location of zero by solving for f(x) = (y2-y1)/(x2-x1) * (x-x1) + y1 = 0
        """
        return x1 - y1 * (x2 - x1) / (y2 - y1)

    wmin = get_zero(w[xmin - 1], I_offset[xmin - 1], w[xmin], I_offset[xmin])
    wmax = get_zero(w[xmax], I_offset[xmax], w[xmax + 1], I_offset[xmax + 1])

    if return_index:
        return abs(wmax - wmin), xmin, xmax
    else:
        return abs(wmax - wmin)


def get_effective_FWHM(w, I):
    """Calculate full-width-at-half-maximum (FWHM) of a triangular
    slit of same area and height 1


    Parameters
    ----------

    w, I: arrays

    Returns
    -------

    fwhm: float
        effective FWHM

    See Also
    --------

    :py:func:`~radis.tools.slit.get_FWHM`,
    :py:func:`~radis.tools.slit.offset_dilate_slit_function`,
    :py:func:`~radis.tools.slit.normalize_slit`,
    :py:func:`~radis.tools.slit.remove_boundary`,
    :py:func:`~radis.tools.slit.plot_slit`,
    :py:func:`~radis.tools.slit.recenter_slit`,
    :py:func:`~radis.tools.slit.crop_slit`

    """

    Imax = I.max()

    area = abs(np.trapz(I, w))

    return area / Imax


def offset_dilate_slit_function(
    w_slit_nm, I_slit, w_nm, slit_dispersion, threshold=0.01, verbose=True
):
    """Offset the slit wavelengths ``w_slit_nm`` to the center of range ``w_nm``, and
    dilate them with ``slit_dispersion(w0)``

    Parameters
    ----------

    w_slit_nm: np.array
        wavelength

        .. warning::
            slit_dispersion is assumed to be in wavelength. Warning if you're
            applying this to an array stored in wavenumbers: don't forget to
            convert.

    I_slit: np.array
        slit intensity

    w_nm: np.array
        wavelength whose center is used to center the slit function

    slit_dispersion: func of (lambda), or ``None``
        spectrometer reciprocal function : dλ/dx(λ)
        If not None, then the slit_dispersion function is used to correct the
        slit function for the whole range. Can be important if slit function
        was measured far from the measured spectrum  (e.g: a slit function
        measured at 632.8 nm will look broader at 350 nm because the spectrometer
        dispersion is higher at 350 nm. Therefore it should be corrected)
        Default ``None``. For more details see :func:`~radis.tools.slit.convolve_with_slit`

    Other Parameters
    ----------------

    threshold: float
        if not ``None``, check that slit dispersion is about constant (< ``threshold`` change)
        on the calculated range. Default 0.01 (1%)

    Returns
    -------

    w_slit, I_slit: np.array
        dilated wavelength (or wavenumber), slit intensity (renormalized)

    See Also
    --------

    :py:func:`~radis.tools.slit.get_FWHM`,
    :py:func:`~radis.tools.slit.get_effective_FWHM`,
    :py:func:`~radis.tools.slit.normalize_slit`,
    :py:func:`~radis.tools.slit.remove_boundary`,
    :py:func:`~radis.tools.slit.plot_slit`,
    :py:func:`~radis.tools.slit.recenter_slit`,
    :py:func:`~radis.tools.slit.crop_slit`

    """
    w0 = w_nm[len(w_nm) // 2]
    wslit0 = w_slit_nm[len(w_slit_nm) // 2]
    slit_disp_0 = slit_dispersion(wslit0)
    range_nm = abs(w_slit_nm[-1] - w_slit_nm[0]) / 2
    w_slit_init = w_slit_nm

    # Check that slit dispersion is about constant (<1% change) on the calculated range
    if threshold:
        if (
            not 1 - threshold
            <= slit_dispersion(
                w_nm.max() - range_nm / slit_disp_0 * slit_dispersion(w_nm.max())
            )
            / slit_dispersion(
                w_nm.min() + range_nm / slit_disp_0 * slit_dispersion(w_nm.min())
            )
            <= 1 + threshold
        ):
            warn(
                "Slit dispersion changes slightly ({2:.2f}%) between {0:.3f} and {1:.3f}nm".format(
                    w_nm.min(),
                    w_nm.max(),
                    abs(slit_dispersion(w_nm.max()) / slit_dispersion(w_nm.min()) - 1)
                    * 100,
                )
                + ". Consider splitting your spectrum",
                SlitDispersionWarning,
            )

    # Offset slit and correct for dispersion
    w_slit_nm = w0 + slit_dispersion(w0) / slit_dispersion(wslit0) * (
        w_slit_nm - wslit0
    )

    if verbose > 2:
        print(
            "{0:.2f} to {1:.2f}nm: slit function FWHM changed from {2:.2f} to {3:.2f}".format(
                wslit0,
                w0,
                get_effective_FWHM(w_slit_init, I_slit),
                get_effective_FWHM(w_slit_nm, I_slit),
            )
        )

    return w_slit_nm, I_slit


def normalize_slit(w_slit, I_slit, norm_by="area"):
    """Normalize slit function with different normalization modes. Warning,
    some change units after convolution!

    Parameters
    ----------

    w_slit, I_slit: np.array
        wavelength and slit intensity

    norm_by: ``'area'``, ``'max'``, or ``None``
        how to normalize. ``'area'`` conserves energy. With ``'max'`` the slit is normalized
        at peak so that the maximum is one.

        .. note::

            ``'max'`` changes the unit of the spectral array, e.g. from
            ``'mW/cm2/sr/µm'`` to ``'mW/cm2/sr')``

        ``None`` doesnt normalize. Default ``'area'``

    Returns
    -------

    w_slit, I_slit: np.array
        wavelength, and normalized slit intensity

    See Also
    --------

    :py:func:`~radis.tools.slit.get_FWHM`,
    :py:func:`~radis.tools.slit.get_effective_FWHM`,
    :py:func:`~radis.tools.slit.offset_dilate_slit_function`,
    :py:func:`~radis.tools.slit.remove_boundary`,
    :py:func:`~radis.tools.slit.plot_slit`,
    :py:func:`~radis.tools.slit.recenter_slit`,
    :py:func:`~radis.tools.slit.crop_slit`

    """

    # Renormalize
    # ---------

    if norm_by == "area":  # normalize by the area
        I_slit = I_slit / abs(np.trapz(I_slit, x=w_slit))
    elif norm_by == "max":  # set maximum to 1
        I_slit = I_slit / abs(np.max(I_slit))
    elif norm_by is None:
        pass
    else:
        raise ValueError("Unknown normalization type: `norm_by` = {0}".format(norm_by))

    return w_slit, I_slit


def remove_boundary(
    w, I_conv, mode, len_I=None, len_I_slit_interp=None, crop_left=None, crop_right=None
):
    """Crop convoluted array to remove boundary effects

    Parameters
    ----------

    w, I_conv: numpy array
        wavelength and intensity of already convoluted quantity
        (ex: ``radiance``)

    mode: ``'valid'``, ``'same'``, ``''crop'``
        ``'same'`` returns output of same length as initial spectra,
        but boundary effects are still visible. ``'valid'`` returns
        output of length len(spectra) - len(slit) + 1, for
        which lines outside of the calculated range have
        no impact. ``'crop'`` just removes ``crop_wings`` points on
        the side.

    Other Parameters
    ----------------

    len_I, len_I_slit_interp: int
        length of initial quantity (ex: ``radiance_noslit``) and length
        of slit function intensity, before convolution.
        Needed to determine the valid range if ``mode='valid'``

    crop_left, crop_right: int
        number of points to discard on each side if ``mode='crop'``
        Values are replaced with ``nan``

    Returns
    -------

    w_conv, I_conv: numpy arrays
        cropped waverange and quantity

    Notes
    -----

    .. note::
        new in 0.9.30 : remove_boundary replaces off-range values with ``nan``
        but keeps the same array size.

    See Also
    --------

    :py:func:`~radis.tools.slit.get_FWHM`,
    :py:func:`~radis.tools.slit.get_effective_FWHM`,
    :py:func:`~radis.tools.slit.offset_dilate_slit_function`,
    :py:func:`~radis.tools.slit.normalize_slit`,
    :py:func:`~radis.tools.slit.plot_slit`,
    :py:func:`~radis.tools.slit.recenter_slit`,
    :py:func:`~radis.tools.slit.crop_slit`

    """

    # Remove boundary effects by adding nans; keep the same x-axis
    if mode == "valid":
        la = min(len_I, len_I_slit_interp)
        a = int((la - 1) / 2)
        b = int((la) / 2)
        # I_conv = I[a:-b]    # former version : we would change the array size
        # w_conv = w[a:-b]
        I_conv[:a] = np.nan
        I_conv[-b:] = np.nan
        w_conv = w
    elif mode == "same":
        I_conv = I_conv
        w_conv = w
    elif mode == "crop":
        if crop_right == 0:
            _crop_right = None
        else:
            _crop_right = -crop_right
        # l = len(I_conv)
        # I_conv = I_conv[crop_left:_crop_right]
        # w_conv = w[crop_left:_crop_right]
        # assert len(I_conv) == l - crop_left - crop_right
        I_conv[:crop_left] = np.nan
        I_conv[_crop_right:] = np.nan
    else:
        raise ValueError("Unexpected mode: {0}".format(mode))

    return w_conv, I_conv


def plot_slit(
    w,
    I=None,
    wunit="",
    plot_unit="same",
    Iunit=None,
    warnings=True,
    ls="-",
    title=None,
    waveunit=None,
):
    """Plot slit, calculate and display FWHM, and calculate effective FWHM.
    FWHM is calculated from the limits of the range above the half width,
    while FWHM is the equivalent width of a triangular slit with the same area


    Parameters
    ----------
    w, I: arrays    or   (str, None)
        if str, open file directly
    waveunit: ``'nm'``, ``'cm-1'`` or ``''``
        unit of input w
    plot_unit: ``'nm'``, ``'cm-1'`` or ``'same'``
        change plot unit (and FWHM units)
    Iunit: str, or None
        give Iunit
    warnings: boolean
        if True, test if slit is correctly centered and output a warning if it
        is not. Default ``True``

    Returns
    -------
    fix, ax: matplotlib objects
        figure and ax

    Examples
    --------

    .. minigallery:: radis.plot_slit

    See Also
    --------

    :py:func:`~radis.tools.slit.get_FWHM`,
    :py:func:`~radis.tools.slit.get_effective_FWHM`,
    :py:func:`~radis.tools.slit.offset_dilate_slit_function`,
    :py:func:`~radis.tools.slit.normalize_slit`,
    :py:func:`~radis.tools.slit.remove_boundary`,
    :py:func:`~radis.tools.slit.recenter_slit`,
    :py:func:`~radis.tools.slit.crop_slit`

    """

    # Deprecated input:
    if waveunit is not None:
        warn(
            "`waveunit=` parameter in convolve_with_slit is now named `wunit=`",
            DeprecationWarning,
        )
        wunit = waveunit

    import matplotlib.pyplot as plt

    from radis.misc.plot import set_style

    set_style()

    try:
        from radis.plot.toolbar import add_tools

        add_tools()  # includes a Ruler to measure slit
    except:
        pass

    # Check input
    if isinstance(w, str) and I is None:
        w, I = np.loadtxt(w).T
    assert len(w) == len(I)
    if anynan(I):
        warn("Slit function has nans")
        w = w[~np.isnan(I)]
        I = I[~np.isnan(I)]
    assert len(I) > 0

    # cast units
    wunit = cast_waveunit(wunit, force_match=False)
    plot_unit = cast_waveunit(plot_unit, force_match=False)
    if plot_unit == "same":
        plot_unit = wunit

    # Convert wavespace unit if needed
    elif wunit == "cm-1" and plot_unit == "nm":  # wavelength > wavenumber
        w = cm2nm(w)
        wunit = "nm"
    elif wunit == "nm" and plot_unit == "cm-1":  # wavenumber > wavelength
        w = nm2cm(w)
        wunit = "cm-1"
    elif wunit == plot_unit:  # same units
        pass
    elif plot_unit == "":  # ???
        pass
    else:
        raise ValueError("Unknown plot unit: {0}".format(plot_unit))

    # Recalculate FWHM
    FWHM, xmin, xmax = get_FWHM(w, I, return_index=True)
    FWHM_eff = get_effective_FWHM(w, I)

    # Get labels
    if plot_unit == "nm":
        xlabel = "Wavelength (nm)"
    elif plot_unit == "cm-1":
        xlabel = "Wavenumber (cm-1)"
    elif plot_unit == "":
        xlabel = "Wavespace"
    else:
        raise ValueError("Unknown unit for plot_unit: {0}".format(plot_unit))
    ylabel = "Slit function"
    if Iunit is not None:
        ylabel += " ({0})".format(Iunit)

    fig, ax = plt.subplots()
    ax.plot(w, I, "o", color="lightgrey")
    ax.plot(
        w,
        I,
        "k",
        ls=ls,
        label="FWHM: {0:.3f} {1}".format(FWHM, plot_unit)
        + "\nEff. FWHM: {0:.3f} {1}".format(FWHM_eff, plot_unit)
        + "\nArea: {0:.3f}".format(abs(np.trapz(I, x=w))),
    )

    # Vertical lines on center, and FWHM
    plt.axvline(w[len(w) // 2], ls="-", lw=2, color="lightgrey")  # center
    plt.axvline(
        w[(xmin + xmax) // 2], ls="--", color="k", lw=0.5
    )  # maximum (should be center)
    plt.axvline(w[xmin], ls="--", color="k", lw=0.5)  # FWHM min
    plt.axvline(w[xmax], ls="--", color="k", lw=0.5)  # FWHM max
    plt.axhline(I.max() / 2, ls="--", color="k", lw=0.5)  # half maximum

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    plt.legend(loc="best", prop={"size": 16})

    if title:
        plt.title(title)

    # extend axis:
    fig.tight_layout()
    xlmin, xlmax = ax.get_xlim()
    ax.set_xlim((xlmin - 0.5, xlmax + 0.5))

    if warnings:
        if w[(xmin + xmax) // 2] != w[len(w) // 2]:
            warn(
                "Slit function doesnt seem centered: center measured with FWHM"
                + " is not the array center (shift: {0:.3f}{1}): This can induce offsets!".format(
                    abs(w[(xmin + xmax) // 2] - w[len(w) // 2]), wunit
                )
            )

        if I[0] != 0 or I[-1] != 0:
            warn("Slit function should have zeros on both sides")

    return fig, ax


def recenter_slit(w_slit, I_slit, verbose=True):
    """Recenters the slit on the maximum calculated from the two
    FWHM limits. To recenter, zeros are added on the shorter side (zero padding)

    See Also
    --------

    :py:func:`~radis.tools.slit.get_FWHM`,
    :py:func:`~radis.tools.slit.get_effective_FWHM`,
    :py:func:`~radis.tools.slit.offset_dilate_slit_function`,
    :py:func:`~radis.tools.slit.normalize_slit`,
    :py:func:`~radis.tools.slit.remove_boundary`,
    :py:func:`~radis.tools.slit.plot_slit`,
    :py:func:`~radis.tools.slit.crop_slit`

    """

    _, xmin, xmax = get_FWHM(w_slit, I_slit, return_index=True)
    xcenter = (xmax + xmin) / 2
    offset = len(w_slit) / 2 - xcenter
    add_zeros = int(2 * offset)
    if add_zeros > 0:
        # add zeros on the left (assume regular step)
        wstep = w_slit[1] - w_slit[0]
        w_left = np.linspace(
            w_slit[0] - wstep * (add_zeros + 1), w_slit[0] - wstep, add_zeros
        )
        w_slit = np.hstack((w_left, w_slit))
        I_slit = np.hstack((np.zeros(add_zeros), I_slit))
        if verbose:
            print("... added zeros to the slit to center it")
    elif add_zeros < 0:
        add_zeros = -add_zeros
        # add zeros on the right (assume regular step)
        wstep = w_slit[-1] - w_slit[-2]
        w_right = np.linspace(
            w_slit[-1] + wstep, w_slit[-1] + wstep * (add_zeros + 1), add_zeros
        )
        w_slit = np.hstack((w_slit, w_right))
        I_slit = np.hstack((I_slit, np.zeros(add_zeros)))
        if verbose:
            print("... added zeros to the slit to center it")
    else:
        pass

    return w_slit, I_slit


def crop_slit(w_slit, I_slit, verbose=True):
    """Removes unnecessary zeros on the side for a faster convolution.
    (remove as many zeros on the left as on the right).

    See Also
    --------

    :py:func:`~radis.tools.slit.get_FWHM`,
    :py:func:`~radis.tools.slit.get_effective_FWHM`,
    :py:func:`~radis.tools.slit.offset_dilate_slit_function`,
    :py:func:`~radis.tools.slit.normalize_slit`,
    :py:func:`~radis.tools.slit.remove_boundary`,
    :py:func:`~radis.tools.slit.plot_slit`,
    :py:func:`~radis.tools.slit.recenter_slit`

    """

    nzeros_index = np.argwhere(I_slit != 0)
    zeros_left = nzeros_index.min()
    zeros_right = len(I_slit) - nzeros_index.max() - 1
    remove = 0
    if zeros_left > 1 and zeros_right > zeros_left:
        remove = zeros_left - 1
    elif zeros_right > 1 and zeros_left > zeros_right:
        remove = zeros_right - 1
    if remove > 0:
        w_slit = w_slit[remove:-remove]
        I_slit = I_slit[remove:-remove]
        if verbose:
            print("... removed {0} zeros to the slit on each side".format(remove))

    return w_slit, I_slit


# %% Slit function models


def import_experimental_slit(
    slit,
    norm_by="area",
    bplot=False,
    wunit="nm",
    auto_recenter=True,
    auto_crop=True,
    center=None,
    scale=1,
    verbose=True,
    waveunit=None,
):
    """Import instrumental slit function and normalize it


    Parameters
    ----------

    slit: str or numpy object (wslit, Islit)
        slit function spectrum.
        If ``str``, the slit is loaded from the corresponding file with::

            w_slit, I_slit = np.loadtxt(slit).T

    norm_by: ``None``, ``'area'``, ``'max'``
        normalisation type. ``'area'`` divides by the area (conserves energy / units
        are unchanged). ``'max'`` divides by the peak value (similar to what is
        done in Specair / changes units). ``None`` does not normalize the
        experimental slit function. Default ``'area'``

    bplot: boolean
        plot normalized slit function (for debugging). Default ``False``

    wunit: ``'nm'``, ``'cm-1'``
        used for plot only. Slit function is generated assuming you use the
        correct wavespace. No conversions are made here. Default ``'nm'``

    Other Parameters
    ----------------

    auto_recenter: bool
        if True, recenters the slit on the maximum calculated from the two
        FWHM limits. To recenter, zeros are added on the shorter side (zero padding)
        Default ``True``

    auto_crop: bool
        If True, remove unnecessary zeros on the side for a faster convolution.
        (remove as many zeros on the left as on the right). Default ``True``

    center: float, None
        normally the slit instrumental slit function is centered on where it
        was measured. If not None, center overwrites the center and
        offsets the slit to the given value. Default ``None``

    scale: float
        multiply slit by an arbitrary factor. Default 1.

    verbose: boolean
        Display messages

    Returns
    -------

    w_slit, I_slit: numpy arrays
        wavelength and intensity of normalized slit function

    See Also
    --------

    :py:func:`~radis.tools.slit.triangular_slit`,
    :py:func:`~radis.tools.slit.trapezoidal_slit`,
    :py:func:`~radis.tools.slit.gaussian_slit`


    """

    # Deprecated input:
    if waveunit is not None:
        warn(
            "`waveunit=` parameter in convolve_with_slit is now named `wunit=`",
            DeprecationWarning,
        )
        wunit = waveunit

    # import
    if isinstance(slit, str):
        try:
            w_slit, I_slit = np.loadtxt(slit).T
        except ValueError:
            w_slit, I_slit = np.loadtxt(slit)
    # numpy input
    elif isinstance(slit, np.ndarray):
        a, b = np.shape(slit)
        if a == 2:
            w_slit, I_slit = slit
        elif b == 2:
            w_slit, I_slit = slit.T
        else:
            raise ValueError("Wrong format of slit_function. Should be 2 x n")
    else:
        raise TypeError("Unexpected type for slit function: {0}".format(type(slit)))

    assert np.shape(w_slit) == np.shape(I_slit)

    # check sides are zeros
    if not I_slit[0] == 0 and I_slit[-1] == 0:
        raise ValueError("Slit function must be null on each side. Fix it")

    # recenter if asked for
    if auto_recenter:
        w_slit, I_slit = recenter_slit(w_slit, I_slit, verbose=verbose)
        # note: if auto_crop is true we may be adding zeros just to remove them
        # right way. Fix that. Someday.

    # remove unnecessary zeros
    if auto_crop:  # (note that we know I_slit has zeros on each side already)
        w_slit, I_slit = crop_slit(w_slit, I_slit, verbose=verbose)

    # Offset slit if needed
    if center is not None:
        w_slit += center - w_slit[len(w_slit) // 2]

    # Normalize
    if norm_by == "area":  # normalize by the area
        #        I_slit /= np.trapz(I_slit, x=w_slit)
        I_slit /= abs(np.trapz(I_slit, x=w_slit))
        Iunit = "1/{0}".format(waveunit)
    elif norm_by == "max":  # set maximum to 1
        I_slit /= abs(np.max(I_slit))
        Iunit = ""
    elif norm_by is None:
        Iunit = None
    else:
        raise ValueError("Unknown normalization type: `norm_by` = {0}".format(norm_by))

    # scale
    I_slit *= scale
    #    if Iunit is not None and scale != 1:
    #        Iunit += 'x{0:.2f}'.format(scale)

    # Plot slit
    if bplot:
        plot_slit(w_slit, I_slit, wunit=wunit, Iunit=Iunit)

    return w_slit, I_slit


def triangular_slit(
    FWHM,
    wstep,
    center=0,
    norm_by="area",
    bplot=False,
    wunit="",
    scale=1,
    footerspacing=0,
    waveunit=None,  # Deprecated
):
    r""" Generate normalized slit function


    Parameters
    ----------

    FWHM: (nm)
        full-width at half maximum

    wstep: (nm)
        wavelength step

    center: (nm)
        center wavelength for the wavelength axis of the slit function

    norm_by: ``'area'``, ``'max'``
        normalisation type. ``'area'`` conserves energy. ``'max'`` is what is
        done in Specair and changes units. Default ``'area'``

    bplot: boolean
        plot normalized slit function (for debugging). Default ``False``

    wunit: '', 'nm', 'cm-1'
        used for plot only. Slit function is generated assuming you use the
        correct wavespace. No conversions are made here.

    scale: float
        multiply slit by an arbitrary factor. Default 1.

    footerspacing: int
        spacing (footer) on left and right. Default 10.


    Returns
    -------

    w, I: numpy arrays
        Slit function of shape::

                       ^
                      / \
                 _ _ /   \ _ _
                :   :  :  :   :
               -a-f -a 0  a  a+f


                with FWHM = (a+1/2)*wstep, `f` spacing on left & right


    Notes
    -----

    slit is generated with an odd number of elements and centered on its
    maximum. This may result in sightly different FWHM if wstep is too large.
    However, slit is corrected so that effective FWHM is preserved.

    See Also
    --------

    :py:func:`~radis.tools.slit.import_experimental_slit`,
    :py:func:`~radis.tools.slit.trapezoidal_slit`,
    :py:func:`~radis.tools.slit.gaussian_slit`

    """

    # Deprecated input:
    if waveunit is not None:
        warn(
            "`waveunit=` parameter in convolve_with_slit is now named `wunit=`",
            DeprecationWarning,
        )
        wunit = waveunit

    # Build first half
    slope = 1 / (FWHM - wstep)
    a = int(FWHM / wstep)
    I = 1 - np.arange(0, a) * slope * wstep  # slope

    # Zeros
    if FWHM % wstep:  # add one extra zero when not a divider
        footerspacing += 1
    f = int(footerspacing)  # add zeros (footer)
    I = np.hstack((I, np.zeros(f)))
    w = np.linspace(0, len(I) * wstep, len(I))

    # Mirror to get second half
    I = np.hstack((I[1:][::-1], I))
    w = np.hstack((-w[1:][::-1], w)) + center

    # Normalize
    if norm_by == "area":  # normalize by the area
        I /= np.trapz(I, x=w)
        Iunit = "1/{0}".format(wunit)
    elif norm_by == "max":  # set maximum to 1
        I /= np.max(I)
        Iunit = ""
    elif norm_by is None:
        Iunit = None
    else:
        raise ValueError("Unknown normalization type: `norm_by` = {0}".format(norm_by))

    # Scale
    I *= scale
    #    if Iunit is not None and scale != 1:
    #        Iunit += 'x{0:.2f}'.format(scale)

    # Plot slit
    if bplot:
        plot_slit(w, I, wunit=wunit, Iunit=Iunit)

    return w, I


# trapezoidal instrumental broadening function of base (nm) and top (nm)


def trapezoidal_slit(
    top,
    base,
    wstep,
    center=0,
    norm_by="area",
    bplot=False,
    wunit="",
    scale=1,
    footerspacing=0,
    waveunit=None,
):
    r""" Build a trapezoidal slit. Remember that FWHM = (top + base) / 2


    Parameters
    ----------
    top: (nm)
        top of the trapeze
    base: (nm)
        base of the trapeze
    wstep: (nm)
        wavelength step

    Other Parameters
    ----------------
    center: (nm)
        center wavelength for the wavelength axis of the slit function
    norm_by: ``'area'``, ``'max'``
        normalisation type. ``'area'`` conserves energy. ``'max'`` is what is
        done in Specair and changes units. Default ``'area'``
    bplot: boolean
        plot normalized slit function (for debugging). Default ``False``
    waveunit: '', 'nm', 'cm-1'
        used for plot only. Slit function is generated assuming you use the
        correct wavespace. No conversions are made here.
    scale: float
        multiply slit by an arbitrary factor. Default 1.
    footerspacing: int
        spacing (footer) on left and right. Default 10.


    Returns
    -------

    w, I: numpy arrays
        Slit function of shape::

                     _______
                    /       \
                   / :     : \
             _ _ _/  :     :  \____
             0    :  :     :  :   0
             :   -b  :     :  b   :
           -b-f     -t     t     b+f


            with `t`and `b` the half-top and half-base in number of elements, `f` spacing
            on left & right. FWHM = (t+b+1)*wstep


    Notes
    -----

    slit is generated with an odd number of elements and centered on its
    maximum. This may result in sightly different FWHM if wstep is too large.
    However, slit is corrected so that effective FWHM is preserved.

    See Also
    --------

    :py:func:`~radis.tools.slit.import_experimental_slit`,
    :py:func:`~radis.tools.slit.triangular_slit`,
    :py:func:`~radis.tools.slit.gaussian_slit`

    """

    if top > base:
        top, base = base, top

    FWHM = (base + top) / 2
    b = 2 * int(top / wstep // 2) + 1  # number of points on top (even)

    # Build first half
    slope = 1 / (FWHM - b * wstep)
    a = int(FWHM / wstep) - b  # number of points in slope
    I = 1 - np.arange(0, a + 1) * slope * wstep  # slope

    if len(I) == 0:
        I = np.ones(1)

    if I[-1] < 0:
        I[-1] = 0
    elif I[-1] > 0:
        # add one extra zero
        footerspacing += 1

    # Zeros
    #    if abs((base-top) % wstep) < wstep:  # add one extra zero when a divider
    #        footerspacing += 1

    f = int(footerspacing)  # add zeros (footer)
    I = np.hstack((I, np.zeros(f)))
    #    w = np.linspace(0, len(I)*wstep, len(I))

    # Mirror to get second half, add top
    I = np.hstack((I[1:][::-1], np.ones(b), I[1:]))
    w = wstep * np.linspace(-len(I) / 2, len(I) / 2, len(I)) + center

    # Normalize
    if norm_by == "area":  # normalize by the area
        I /= np.trapz(I, x=w)
        Iunit = "1/{0}".format(wunit)
    elif norm_by == "max":  # set maximum to 1
        I /= np.max(I)
        Iunit = ""
    elif norm_by is None:
        Iunit = None
    else:
        raise ValueError("Unknown normalization type: `norm_by` = {0}".format(norm_by))

    # scale
    I *= scale
    #    if Iunit is not None and scale != 1:
    #        Iunit += 'x{0:.2f}'.format(scale)

    # Plot slit
    if bplot:
        plot_slit(w, I, wunit=wunit, Iunit=Iunit)

    return w, I


def gaussian_slit(
    FWHM,
    wstep,
    center=0,
    norm_by="area",
    bplot=False,
    wunit="",
    calc_range=4,
    scale=1,
    footerspacing=0,
    waveunit=None,  # Deprecated
):
    r""" Generate normalized slit function


    Parameters
    ----------
    FWHM: (nm)
        full-width at half maximum
    wstep: (nm)
        wavelength step

    Other Parameters
    ----------------
    center: (nm)
        center wavelength for the wavelength axis of the slit function
    norm_by: ``'area'``, ``'max'``
        normalisation type. ``'area'`` conserves energy. ``'max'`` is what is
        done in Specair and changes units. Default ``'area'``
    bplot: boolean
        plot normalized slit function (for debugging). Default ``False``
    wunit: '', 'nm', 'cm-1'
        used for plot only. Slit function is generated assuming you use the
        correct wavespace. No conversions are made here.
    calc_range: float   (number of sigma)
        function broadening range expressed in number of standard deviation
    scale: float
        multiply slit by an arbitrary factor. Default 1.
    footerspacing: int
        spacing (footer) on left and right. Default 10


    Returns
    -------
    w, I: numpy arrays
        Slit function of shape::

                     .-.
                    /   \
                   /     \
             _ _.-'       `-._ _
             0    [gaussian]   0
             :  :     :     :  :
           -c-f -c    0     c  c+f          (c : calc_range)


            with `c` cutoff in number of elements, `f` spacing on left & right


    Notes
    -----

    slit is generated with an odd number of elements and centered on its
    maximum. This may result in sightly different FWHM if wstep is too large.

    See Also
    --------

    :py:func:`~radis.tools.slit.import_experimental_slit`,
    :py:func:`~radis.tools.slit.triangular_slit`,
    :py:func:`~radis.tools.slit.trapezoidal_slit`


    """

    f = int(footerspacing)  # spacing (footer) on left and right
    sigma = FWHM / 2 / sqrt(2 * ln(2))

    # half-base in number of elements (even number)
    a = 2 * int(calc_range * sigma // wstep / 2)

    # 2 sigma: gaussian non calculated residual: 5%
    # 3 sigma: gaussian non calculated residual: 1%

    w0 = wstep * np.linspace(-a, a, 2 * a + 1)  # centered
    Igauss = exp(-(w0**2) / (2 * sigma**2))

    I = np.hstack((np.zeros(f), Igauss, np.zeros(f)))
    w = wstep * np.linspace(-(a + f), (a + f), 2 * a + 2 * f + 1) + center

    # Normalize
    if norm_by == "area":  # normalize by the area
        I /= np.trapz(I, x=w)
        Iunit = "1/{0}".format(wunit)
    elif norm_by == "max":  # set maximum to 1
        I /= np.max(I)
        Iunit = ""
    elif norm_by is None:
        Iunit = None
    else:
        raise ValueError("Unknown normalization type: `norm_by` = {0}".format(norm_by))

    # scale
    I *= scale
    #    if Iunit is not None and scale != 1:
    #        Iunit += 'x{0:.2f}'.format(scale)

    # Plot slit
    if bplot:
        plot_slit(w, I, wunit=wunit, Iunit=Iunit)

    return w, I


# %% Test
if __name__ == "__main__":
    from radis.test.tools.test_slit import _run_testcases

    print("Testing slit.py: ", _run_testcases())



# -*- coding: utf-8 -*-
"""
Created on Wed Jul  7 12:55:28 2021

@author: erwan

Test the :py:func:`~radis.tools.gascomp.get_eq_mole_fraction` function.
"""


def test_get_eq_mole_fraction(*args, **kwargs):
    from radis.misc.basics import all_in
    from radis.tools.gascomp import get_eq_mole_fraction

    gas = get_eq_mole_fraction("CO2:1", 3000, 1e5)

    assert all_in(["C", "CO", "CO2", "O", "O2"], gas.keys())

    return True


if __name__ == "__main__":
    test_get_eq_mole_fraction()



import atexit
import inspect
import sys
from functools import partial
from io import StringIO


class CodeProfiler:
    def __init__(
        self,
        test_func=None,
        whitelist={"radis"},  # Look for these words in the file path.
        exclusions={"<"},  # Ignore <listcomp>, etc. in the function name.
        filename="code_profiler_out.txt",
    ):

        self.dbg_out = StringIO()
        self.test_func = test_func
        sys.setprofile(self.tracefunc)
        self.test_func = test_func
        self.memorized = set()
        self.stack_level = 0
        self.white_list = whitelist
        self.exclusions = exclusions
        atexit.register(partial(self.dump, filename=filename))

    def tracefunc(self, frame, event, arg):

        if event == "call":
            self.stack_level += 1

            unique_id = frame.f_code.co_filename + str(frame.f_lineno)
            if unique_id in self.memorized:
                return

            # Part of filename MUST be in white list.
            if any(x in frame.f_code.co_filename for x in self.white_list) and not any(
                x in frame.f_code.co_name for x in self.exclusions
            ):

                if "self" in frame.f_locals:
                    module_name = frame.f_locals["self"].__class__.__module__
                    class_name = frame.f_locals["self"].__class__.__name__
                    func_name = (
                        module_name + "." + class_name + "." + frame.f_code.co_name
                    )
                else:
                    module_name = inspect.getmodulename(frame.f_code.co_filename) + "."
                    func_name = module_name + frame.f_code.co_name

                func_name = self.stack_level * " ." + " " + func_name
                txt = "{: <60} # {}, {}\n".format(
                    func_name, frame.f_code.co_filename, frame.f_lineno
                )
                self.dbg_out.write(txt)
                self.memorized.add(unique_id)

        elif event == "return":
            if self.test_func != None:
                if self.test_func(frame):
                    self.dbg_out.write("\n\n~~~ TEST FUNCTION SUCCESS!!! ~~~\n\n")
                    self.test_func = None

            self.stack_level -= 1

    def dump(self, filename="code_profiler_out.txt"):
        ptr = self.dbg_out.tell()
        self.dbg_out.seek(0)
        with open(filename, "w") as f:
            f.write(self.dbg_out.read())
        self.dbg_out.seek(ptr)



# -*- coding: utf-8 -*-
"""

Predefined molecules using RADIS built-in
:ref:`spectroscopic constants <label_db_spectroscopic_constants>`

See :py:data:`~radis.db.molecules.Molecules`

-------------------------------------------------------------------------------


"""

from radis.db.classes import ElectronicState
from radis.phys.convert import eV2cm

# %% Define some commonly used molecules


# CO
# --

# Define with default diatomic constants
CO_X_iso1 = ElectronicState(
    "CO",
    isotope=1,
    state="X",
    term_symbol="1Σ+",
    spectroscopic_constants="default",
    spectroscopic_constants_type="dunham",
    vmax=17,  # max level for Dunham's expansion
    vmax_morse=48,
    Ediss=eV2cm(11.16),
)
"""CO first isotope (:math:`^{16}O^{12}C`), electronic ground state"""
CO_X_iso2 = ElectronicState(
    "CO",
    isotope=2,
    state="X",
    term_symbol="1Σ+",
    spectroscopic_constants="default",
    spectroscopic_constants_type="dunham",
    vmax=17,  # max level for Dunham's expansion
    vmax_morse=48,
    Ediss=eV2cm(11.16),
)
"""CO 2nd isotope (:math:`^{16}O^{13}C`), electronic ground state"""
CO_X_iso3 = ElectronicState(
    "CO",
    isotope=3,
    state="X",
    term_symbol="1Σ+",
    spectroscopic_constants="default",
    spectroscopic_constants_type="dunham",
    vmax=17,  # max level for Dunham's expansion
    vmax_morse=48,
    Ediss=eV2cm(11.16),
)
"""CO 3rd isotope (:math:`^{18}O^{12}C`) electronic ground state"""

# CO2
# ---

CO2_X_626 = ElectronicState(
    "CO2",
    isotope=1,
    state="X",
    term_symbol="1Σu+",
    spectroscopic_constants="default",
    spectroscopic_constants_type="herzberg",
    Ediss=44600,
)
"""CO2 1st isotope (:math:`^{16}O^{12}C^{16}O`), electronic ground state"""
CO2_X_636 = ElectronicState(
    "CO2",
    isotope=2,
    state="X",
    term_symbol="1Σu+",
    spectroscopic_constants="default",
    spectroscopic_constants_type="herzberg",
    Ediss=44600,
)
"""CO2 2nd isotope (:math:`^{16}O^{13}C^{16}O`), electronic ground state"""
CO2_X_628 = ElectronicState(
    "CO2",
    isotope=3,
    state="X",
    term_symbol="1Σu+",
    spectroscopic_constants="default",
    spectroscopic_constants_type="herzberg",
    Ediss=44600,
)
"""CO2 3rd isotope (:math:`^{16}O^{12}C^{18}O`), electronic ground state"""
CO2_X_627 = ElectronicState(
    "CO2",
    isotope=4,
    state="X",
    term_symbol="1Σu+",
    spectroscopic_constants="default",
    spectroscopic_constants_type="herzberg",
    Ediss=44600,
)
"""CO2 4th isotope (:math:`^{16}O^{12}C^{17}O`), electronic ground state"""


# %% Dictionary of predefined molecules
# -----------

#  Molecule  Isotope  ElecState
#         :      :       :
Molecules = {
    "CO": {
        1: {"X": CO_X_iso1},
        2: {"X": CO_X_iso2},
        3: {"X": CO_X_iso3},
    },
    "CO2": {
        1: {"X": CO2_X_626},
        2: {"X": CO2_X_636},
        3: {"X": CO2_X_628},
        4: {"X": CO2_X_627},
    },
}
"""dict: list of Electronic states whose energy levels can be calculated with RADIS
built-in :ref:`spectroscopic constants <label_db_spectroscopic_constants>`.

For references refer to the definition of each molecule:
CO:

- :py:data:`~radis.db.molecules.CO_X_iso1`
- :py:data:`~radis.db.molecules.CO_X_iso2`
- :py:data:`~radis.db.molecules.CO_X_iso3`

CO2:

- :py:data:`~radis.db.molecules.CO2_X_626`
- :py:data:`~radis.db.molecules.CO2_X_636`
- :py:data:`~radis.db.molecules.CO2_X_628`
- :py:data:`~radis.db.molecules.CO2_X_627`

See Also
--------

:py:func:`~radis.db.molecules.getMolecule`
:py:class:`~radis.db.classes.ElectronicState`

"""


def getMolecule(
    molecule, isotope=None, electronic_state=None, verbose=True
) -> ElectronicState:
    """Get an :py:class:`~radis.db.classes.ElectronicState` object in the
    RADIS :py:data:`~radis.db.molecules.Molecules` list, which use the defaults
    :ref:`spectroscopic constants <label_db_spectroscopic_constants>`.

    Parameters
    ----------
    molecule: str
        molecule name
    isotope: int, or ``None``
        isotope number. if None, only one isotope must exist in database. Else,
        an error is raised
    electronic_state: str
        if None, only one electronic state must exist in database. Else, an error
        is raised
    verbose: boolean
        if ``True``, print which electronic state we got

    Returns
    -------
    ElectronicState: an :py:class:`~radis.db.classes.ElectronicState` object.

    Examples
    --------
    Get rovibrational energies using the default :ref:`spectroscopic constants <label_db_spectroscopic_constants>`:
    ::

        from radis import getMolecule

        # Here we get the energy of the v=6, J=3 level of the 2nd isotope of CO:

        CO = getMolecule("CO", 2, "X")
        print(CO.Erovib(6, 3))

    .. minigallery:: radis.getMolecule

    See Also
    --------
    :py:data:`~radis.db.molecules.Molecules`,
    :py:class:`~radis.db.classes.ElectronicState`

    """

    # molecule
    try:
        mol = Molecules[molecule]
    except KeyError as err:
        raise KeyError(
            "{0} is not defined in molecules with built-in ".format(molecule)
            + "spectroscopic constants. Choose one of: {0}".format(
                list(Molecules.keys())
            )
        ) from err

    # isotope
    if isotope is None:
        if len(list(mol.keys())) != 1:
            raise ValueError(
                "Please precise which isotope among: {0}".format(list(mol.keys()))
            )
        isotope = list(mol.keys())[0]
    try:
        iso = mol[isotope]
    except KeyError as err:
        raise KeyError(
            "Isotope {0} is not defined for molecule {1}. Choose one of: {2}".format(
                isotope, molecule, list(mol.keys())
            )
        ) from err

    # electronic state
    if electronic_state is None:
        if len(list(iso.keys())) != 1:
            raise ValueError(
                "Please choose which electronic state among: {0}".format(
                    list(iso.keys())
                )
            )
        electronic_state = list(iso.keys())[0]
    try:
        state = iso[electronic_state]
    except KeyError as err:
        raise KeyError(
            "{0} is not defined for molecule {1}(iso={2}). Choose one of: {3}".format(
                electronic_state, molecule, isotope, list(mol.keys())
            )
        ) from err

    # print name
    if verbose >= 2:
        print(r"Found {0} in RADIS database".format(state.get_fullname()))

    # Return
    return state


# %% Test


if __name__ == "__main__":

    from radis.test.db.test_molecules import _run_testcases

    print(("Testing molecules.py", _run_testcases()))



"""Install file for RADIS.

Typical install procedure, plus:

- auto-convert README.rst to long_description, removing some sphinx-only syntax
so it can be rendered by PyPi
- read version number from __version__.txt
- some dependencies should be installed manually (typically: all packages with
compiled components such as numpy, pandas, etc.)


Examples
--------

Install (normal, use-only)::

    python setup.py install

Or (create an alias, so you can still edit)::

    python setup.py develop

Notes
-----

For developers:

when creating a new version, just update the __version__.txt file

to register it on Pypi see register.py::

    python register.py
"""
import io
import re
import sys
from os.path import abspath, dirname, exists, join

# from setuptools import Extension, find_packages, setup #for cython - unused after v0.15
from setuptools import setup

# Build description from README (PyPi compatible)
# -----------------------------------------------

# Utils to format RST
def yield_sphinx_only_markup(lines):
    """Cleans-up Sphinx-only constructs (ie from README.rst), so that *PyPi*
    can format it properly.

    To check for remaining errors, install ``sphinx`` and run::

        python setup.py --long-description | sed -file 'this_file.sed' | rst2html.py  --halt=warning

    :param file_inp:     a `filename` or ``sys.stdin``?
    :param file_out:     a `filename` or ``sys.stdout`?`

    References
    ----------

    https://stackoverflow.com/questions/16367770/my-rst-readme-is-not-formatted-on-pypi-python-org

    Notes
    -----

    Check output with::

        python setup.py --long-description | rst2html.py > output.html
    """
    substs = [
        ## Selected Sphinx-only Roles.
        #
        (r":abbr:`([^`]+)`", r"\1"),
        (r":ref:`([^`]+)`", r"`\1`_"),
        (r":term:`([^`]+)`", r"**\1**"),
        (r":dfn:`([^`]+)`", r"**\1**"),
        (r":(samp|guilabel|menuselection):`([^`]+)`", r"``\2``"),
        ## Sphinx-only roles:
        #        :foo:`bar`   --> foo(``bar``)
        #        :a:foo:`bar` XXX afoo(``bar``)
        #
        # (r'(:(\w+))?:(\w+):`([^`]*)`', r'\2\3(``\4``)'),
        (r":(\w+):`([^`]*)`", r"\1(``\2``)"),
        ## Sphinx-only Directives.
        #
        (r"\.\. doctest", r"code-block"),
        (r"\.\. plot::", r".. "),
        (r"\.\. seealso", r"info"),
        (r"\.\. glossary", r"rubric"),
        (r"\.\. figure::", r".. "),
        ## Other
        #
        (r"\|version\|", r"x.x.x"),
        ## added to make RADIS docs Pypi compatible
        #        (r'\.\. image::',          r'.. '),
        #        (r'\.\. |CO2| replace:: CO\ :sub:`2`',          r'.. '),
    ]

    regex_subs = [(re.compile(regex, re.IGNORECASE), sub) for (regex, sub) in substs]

    def clean_line(line):
        try:
            for (regex, sub) in regex_subs:
                line = regex.sub(sub, line)
        except Exception as ex:
            print(("ERROR: %s, (line(%s)" % (regex, sub)))
            raise ex

        return line

    for line in lines:
        yield clean_line(line)


# (note: README.rst has been converted to README.md by register.py, and cleaned afterwards )
description = "A fast line-by-line code for high-resolution infrared molecular spectra"
readme_path = join(abspath(dirname(__file__)), "README.md")
if not exists(readme_path):
    long_description = description
else:
    with io.open(readme_path, encoding="utf-8") as f:
        long_description = f.read()

# Read version number from file
with open(join(dirname(__file__), "radis", "__version__.txt")) as version_file:
    __version__ = version_file.read().strip()


# %% Cython extensions:
# Ensures RADIS still builds if without Cython


class BuildFailed(Exception):
    pass


# from distutils.errors import CCompilerError, DistutilsExecError, DistutilsPlatformError #for cython - unused after v0.15


def show_message(*lines):
    """Note : will only happen if user installs with `pip install -v`"""
    print("=" * 74, file=sys.stderr)
    for line in lines:
        print(line, file=sys.stderr)
    print("=" * 74, file=sys.stderr)


# Deprecated - See Main install routine for radis<0.15
# def get_ext_modules(with_binaries):
#     """
#     Parameters
#     ----------
#     with_binaries: bool
#         if False, do not try to build Cython extensions
#     """
#     # Based on code from https://github.com/pallets/markupsafe/blob/main/setup.py

#     print(sys.version)
#     ext_modules = []
#     cmdclass = {}

#     if not with_binaries:
#         # skip building extensions
#         return {"cmdclass": cmdclass, "ext_modules": ext_modules}

#     # TO-DO: set language level

#     #### Cython is now completely removed - August 2024 - Radis 0.15 ###
#     # try:
#     #     import cython
#     # except (ModuleNotFoundError) as err:
#     #     raise BuildFailed(
#     #         "Cython not found : Skipping all Cython extensions...!"
#     #     ) from err

#     # print("Cython " + cython.__version__)

#     # from Cython.Distutils import build_ext

#     # class build_ext_subclass(build_ext):
#     #     def build_extensions(self):
#     #         c = self.compiler.compiler_type
#     #         copt = {
#     #             "msvc": ["/openmp", "/Ox", "/fp:fast", "/favor:INTEL64"],
#     #             "mingw32": ["-fopenmp", "-O3", "-ffast-math", "-march=native"],
#     #         }

#     #         lopt = {"mingw32": ["-fopenmp"]}

#     #         print("Compiling with " + c + "...")
#     #         try:
#     #             for e in self.extensions:
#     #                 e.extra_compile_args = copt[c]
#     #         except (KeyError):
#     #             pass
#     #         try:
#     #             for e in self.extensions:
#     #                 e.extra_link_args = lopt[c]
#     #         except (KeyError):
#     #             pass
#     #         try:
#     #             build_ext.build_extensions(self)
#     #         except (CCompilerError, DistutilsExecError, DistutilsPlatformError) as err:
#     #             raise BuildFailed() from err

#     # from numpy import get_include

#     # ext_modules.append(
#     #     Extension(
#     #         "radis_cython_extensions",
#     #         sources=[
#     #             "./radis/cython/radis_cython_extensions.pyx",
#     #         ],
#     #         include_dirs=[get_include()],
#     #         language="c++",
#     #         extra_link_args=[],
#     #         define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
#     #     )
#     # )

#     # cmdclass["build_ext"] = build_ext_subclass

#     return {"cmdclass": cmdclass, "ext_modules": ext_modules}


#%% Main install routine
def run_setup():
    setup(
        name="radis",
        version=__version__,
        description=description,
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/radis/radis",
        author="Erwan Pannier",
        author_email="erwan.pannier@gmail.com",
        license="GNU Lesser General Public License v3 (LGPLv3)",
        keywords=[
            "spectrum",
            "infrared",
            "spectra",
            "radiation",
            "nonequilibrium",
            "spectroscopy",
            "molecules",
            "HITRAN",
            "hitemp",
            "exomol",
            "line-by-line",
        ],
        # packages=find_packages(), #this misses the gpu folder, see https://github.com/radis/radis/issues/681
        install_requires=[
            # Although it seems like a duplicate of requirements.txt, it is NECESSARY to tell pip what are the dependencies, issue #680.
            # (requirements.txt is only for the user, not for pip, see https://packaging.python.org/discussions/install-requires-vs-requirements/ and https://stackoverflow.com/questions/14399534/reference-requirements-txt-for-the-install-requires-kwarg-in-setuptools-setup-py)
            ### From environment.yml
            "astropy>=4.3.1",  # Unit aware calculations
            "astroquery>=0.4.6",  # to fetch HITRAN databases
            "beautifulsoup4>=4.10.0",  # parse ExoMol website
            "cantera>=2.5.1",  # for chemical equilibrium computations
            "configparser",
            "habanero>=1.2.0",  # CrossRef API to retrieve data from doi
            "h5py>=3.2.1",  # load HDF5
            "joblib",  # for parallel loading of SpecDatabase
            "lmfit",  # for new fitting modules
            "matplotlib",
            "numpy",
            "numba",  # just-in-time compiler
            "pandas",
            "plotly>=2.5.1",  # for line survey HTML output
            "psutil",  # to get user RAM
            "tables",  # for pandas to HDF5 export - WARNING named "pytables" in conda
            "scipy>=1.4.0",
            "seaborn",  # other matplotlib themes
            "termcolor",  # terminal colors
            "specutils",
            ### From requirements.txt
            "lxml",  # parser used for ExoMol website
            "hjson",  # Json with comments (for default_radis.json)
            "publib",  # Plotting styles for Matplotlib. Version, update to 0.4.0 needed for matplotlib==3.8. However, only 0.3.2 is compatible with python==3.8 (see #647)
            "hitran-api",  # HAPI, used to access TIPS partition functions
            "peakutils",
            "ruamel.yaml",
            "json-tricks>=3.15.0",  # to deal with non jsonable formats
            "mpldatacursor",
            'nvidia-cufft-cu11; sys_platform != "darwin" ',
            "periodictable",
            'vaex-core ; python_version < "3.11"',
            'vaex-hdf5 ; python_version < "3.11"',
            'vaex-viz ; python_version < "3.11"',
        ],
        extras_require={
            "dev": [
                "numpydoc",  # for Jedi (autocompletion) to recognize
                "black>=20.8b1",  # for code-linting in accordance to PEP8
                "isort",  # for sorting imports
                "pre-commit",  # to enforce Black before each commit
                "pytest",  # to run test suite
                "ipython>=7.0.0",  # useful for fast debugging
            ],
            "docs": [
                "sphinx-autodoc-annotation",
                "sphinx_autodoc_defaultargs>=0.1.2",
                "sphinx>=1.7.0",
                "astroquery>=0.3.9",
                "sphinxcontrib-apidoc",
                "sphinx-gallery",
                "lmfit",
                "pytest",  # Sphinx autodoc also parses test suite
            ],
        },
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
            "Topic :: Scientific/Engineering",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.8",  # end of life, Novembre 2024, https://devguide.python.org/versions/
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Operating System :: OS Independent",
        ],
        # **get_ext_modules(with_binary), #see Main install routine for radis<0.15
        include_package_data=True,  # add non .py data files in MANIFEST.in
        # package_data={'radis': ['radis/phys/units.txt']},
        zip_safe=False,  # impossible as long as we have external files read with __file__ syntax
        platforms="any",
    )


# %% Run Main install routine
run_setup()
# %% Run Main install routine for radis<0.15 - Cython is not implemented anymore
# try:
#     run_setup(with_binary=False)
# except BuildFailed:
#     import traceback

#     traceback.print_exc()
#     show_message(
#         "WARNING: RADIS C-extension could not be compiled, speedups",
#         " are not enabled.",
#         "Failure information, if any, is above.",
#         "Retrying the build without the C extension now.",
#     )
#     run_setup(with_binary=False)
#     show_message(
#         "WARNING: RADIS C-extension could not be compiled, speedups"
#         " are not enabled.",
#         "Plain-Python build succeeded.",
#     )



# -*- coding: utf-8 -*-
"""Conversion formulas between different units (works with numpy arrays)

Checks the order of magnitudes to help detect conversion errors (comparing with
standard order of magnitudes in plasma physics)

Examples
--------

Get equivalent width in nm of a 10cm-1 width at 380 nm

    >>> from radis import *
    >>> dcm2dnm(10, nm2cm(380))


-------------------------------------------------------------------------------
"""

import numpy as np

from radis.phys.air import air2vacuum, vacuum2air
from radis.phys.constants import c, eV, h, hc_k, k_b

# Make conv2 accessible from .convert:
from radis.phys.units import conv2

assert conv2  # removes flake8 warnings

# %% Energy units


def J2eV(E):
    """J to eV."""
    return E / eV


def J2K(E):
    """J to Kelvin."""
    return E / k_b


def J2cm(E):
    """J to cm-1."""
    return E / (h * c) / 100


def eV2J(E):
    """eV to J."""
    _asserteV(E)
    return E * eV


def eV2nm(E):
    """nm to eV."""
    _asserteV(E)
    return 1 / E * 1e9 * (h * c) / eV


def eV2K(E):
    """eV to K."""
    _asserteV(E)
    return E * eV / k_b


def eV2cm(E):
    """eV to cm."""
    _asserteV(E)
    return E * eV / (h * c) / 100


def K2eV(E):
    """eV to K."""
    _assertK(E)
    return E * k_b / eV


def K2J(E):
    """Kelvin to J."""
    _assertK(E)
    return E * k_b


def K2cm(E):
    """K to cm-1."""
    _assertK(E)
    return E / hc_k


def cm2J(E):
    """cm-1 to J."""
    _assertcm(E)
    return (E * 100) * (h * c)


def cm2J_vaex(E):
    """cm-1 to J."""
    _assertcm_vaex(E)
    return (E * 100) * (h * c)


def cm2K(E):
    """cm-1 to K.

    That's the classical 1.44 K/cm-1. See
    :data:`~radis.phys.constants.hc_k`
    """
    _assertcm(E)
    return E * hc_k


def cm2eV(E):
    """cm-1 to eV."""
    _assertcm(E)
    return (E * 100) * (h * c) / eV


# %% Wavelength, wavenumbers and frequencies


def cm2nm(wl_cm1):
    """cm-1 to (vacuum) nm."""
    return 1 / wl_cm1 * 1e9 / 100


def cm2hz(wl_cm1):
    """wavenumber to frequency wl_cm1, output in hz."""
    return wl_cm1 * c * 100


def nm2cm(wl_nm):
    """(vacuum) nm to cm-1."""
    return 1 / wl_nm * 1e9 / 100


def cm2nm_air(wl_cm1):
    """cm-1 to (air) nm.

    References
    ----------

    :func:`~radis.phys.air.vacuum2air'
    """
    return vacuum2air(cm2nm(wl_cm1))


def nm_air2cm(wl_nm_air):
    """(air) nm to cm-1.

    References
    ----------

    :func:`~radis.phys.air.air2vacuum'
    """
    return nm2cm(air2vacuum(wl_nm_air))


def nm2eV(wl_nm):
    """nm to eV."""
    return 1 / wl_nm * 1e9 * (h * c) / eV


def hz2nm(f_Hz):
    """frequency to wavelength f in Hz, output in nm."""
    return c * 1e9 / f_Hz


def hz2cm(f_Hz):
    """frequency to wavenumber f in Hz, output in cm-1."""
    return f_Hz / c / 100


def nm2hz(lbd_nm):
    """wavelength to frequency lbd in Hz, output in nm."""
    return c * 1e9 / lbd_nm


# Function for safe division


def zero2nan(x):
    try:
        x[x == 0] = np.nan
        return x
    except (TypeError):
        return x if x else np.nan


def div_safe(f):
    return lambda x: f(zero2nan(x))


# Convert Broadenings


def dcm2dnm(delta_nu, nu_0):
    """Converts (ex: FWHM) from Δcm to Δnm.

    Parameters
    ----------

    delta_nu: float (cm-1)
        wavenumber broadening

    nu_0: float (cm-1)
        center wavenumber

    Returns
    -------

    delta_nm: float (nm)
        broadening in wavelength (vacuum)
    """
    return cm2nm(nu_0 - delta_nu / 2) - cm2nm(nu_0 + delta_nu / 2)


def dnm2dcm(delta_lbd, lbd_0):
    """Converts (ex: FWHM) from Δnm to Δcm.

    Parameters
    ----------

    delta_lbd: float (nm)
        wavelength broadening (vacuum)

    lbd_0: float (nm)
        center wavelength (vacuum)

    Returns
    -------

    delta_cm: float (cm-1)
        broadening in wavenumber
    """
    return nm2cm(lbd_0 - delta_lbd / 2) - nm2cm(lbd_0 + delta_lbd / 2)


def dcm2dnm_air(delta_nu, nu_0):
    """Converts (ex: FWHM) from Δcm to Δnm.

    Parameters
    ----------

    delta_nu: float (cm-1)
        wavenumber broadening

    nu_0: float (cm-1)
        center wavenumber

    Returns
    -------

    delta_nm: float (nm)
        broadening in wavelength (air)
    """
    return cm2nm_air(nu_0 - delta_nu / 2) - cm2nm_air(nu_0 + delta_nu / 2)


def dnm_air2dcm(delta_lbd, lbd_0):
    """Converts (ex: FWHM) from Δnm to Δcm.

    Parameters
    ----------

    delta_lbd: float (nm)
        wavelength broadening (air)

    lbd_0: float (nm)
        center wavelength (air)

    Returns
    -------

    delta_cm: float (cm-1)
        broadening in wavenumber
    """
    return nm_air2cm(lbd_0 - delta_lbd / 2) - nm_air2cm(lbd_0 + delta_lbd / 2)


def dhz2dnm(deltaf_hz, f_0):
    """Converts (ex: FWHM) from ΔHz to Δnm.

    Parameters
    ----------

    deltaf_hz: Hz
        frequency broadening

    nu_0: Hz
        center frequency
    """
    return hz2nm(f_0 - deltaf_hz / 2) - hz2nm(f_0 + deltaf_hz / 2)


def dnm2dhz(delta_lbd, lbd_0):
    """Converts (ex: FWHM) from Δnm to Δhz.

    Parameters
    ----------

    delta_lbd: nm
        wavelength broadening

    lbd_0: nm
        center wavelength
    """
    return nm2hz(lbd_0 - delta_lbd / 2) - nm2hz(lbd_0 + delta_lbd / 2)


# %% Pressure units


def torr2bar(p_torr):
    """Torr to bar."""
    return p_torr * 1.01325 / 760


def torr2atm(p_atm):
    return p_atm / 760


def bar2torr(p_bar):
    return p_bar / 1.01325 * 760


def bar2atm(p_bar):
    return p_bar / 1.01325


def atm2torr(p_atm):
    return p_atm * 760


def atm2bar(p_atm):
    return p_atm * 1.01325


# %% Assert functions


def _magn(x):
    return np.round((np.log10(np.abs(x))))


def _assertK(E):
    if np.sum(np.abs(E)) != 0:  # check E != 0 for both floats and arrays
        try:
            m = _magn(E)
            assert ((0 <= m) & (m <= 6)).all()
        except AssertionError:
            print(("Warning. Input values may not be in Kelvin", E, "K?"))


def _assertcm(E):
    if np.sum(np.abs(E)) != 0:  # check E != 0 for both floats and arrays
        try:
            m = _magn(E)
            assert ((1 <= m) & (m <= 5)).all()
        except AssertionError:
            print(("Warning. Input values may not be in cm-1", E, "cm-1?"))


def _assertcm_vaex(E):
    if E.abs().sum():  # check E != 0 for both floats and arrays
        try:
            m = _magn(E)
            assert (((1 <= m) & (m <= 5))).unique() == [True]
        except AssertionError:
            print(("Warning. Input values may not be in cm-1", E, "cm-1?"))


def _asserteV(E):
    if np.sum(np.abs(E)) != 0:  # check E != 0 for both floats and arrays
        try:
            m = _magn(E)
            assert ((0 <= m) & (m <= 2)).all()
        except AssertionError:
            print(("Warning. Input values may not be in eV", E, "eV?"))


# %% Test
if __name__ == "__main__":
    from radis.test.test_phys import test_convert

    print(("Test :", test_convert()))
